#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os, random
random.seed(int(os.getenv("SEED"), 16))

import json

from prjxray import util
from prjxray import verilog
from prjxray.db import Database

# =============================================================================


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    tile_list = []
    for tile_name in sorted(grid.tiles()):
        if "IOB33" not in tile_name or "SING" in tile_name:
            continue
        tile_list.append(tile_name)

    get_xy = util.create_xy_fun('[LR]IOB33_')
    tile_list.sort(key=get_xy)

    for iob_tile_name in tile_list:
        iob_gridinfo = grid.gridinfo_at_loc(
            grid.loc_of_tilename(iob_tile_name))

        iob33s = [k for k, v in iob_gridinfo.sites.items() if v == "IOB33S"][0]
        iob33m = [k for k, v in iob_gridinfo.sites.items() if v == "IOB33M"][0]

        top_sites = {
            "IOB": iob33m,
            "ILOGIC": iob33m.replace("IOB", "ILOGIC"),
            "IDELAY": iob33m.replace("IOB", "IDELAY"),
        }

        bot_sites = {
            "IOB": iob33s,
            "ILOGIC": iob33s.replace("IOB", "ILOGIC"),
            "IDELAY": iob33s.replace("IOB", "IDELAY"),
        }

        yield iob_tile_name, top_sites, bot_sites


def gen_iserdes(loc):

    # Site params
    params = {
        "SITE_LOC": verilog.quote(loc),
        "USE_IDELAY": random.randint(0, 1),
        "BEL_TYPE": verilog.quote("ISERDESE2"),
        "INIT_Q1": random.randint(0, 1),
        "INIT_Q2": random.randint(0, 1),
        "INIT_Q3": random.randint(0, 1),
        "INIT_Q4": random.randint(0, 1),
        "SRVAL_Q1": random.randint(0, 1),
        "SRVAL_Q2": random.randint(0, 1),
        "SRVAL_Q3": random.randint(0, 1),
        "SRVAL_Q4": random.randint(0, 1),
        "NUM_CE": random.randint(1, 2),

        # The following one shows negative correlation (0 - not inverted)
        "IS_D_INVERTED": random.randint(0, 1),

        # No bits were found for parameters below
        "IS_OCLKB_INVERTED": random.randint(0, 1),
        "IS_OCLK_INVERTED": random.randint(0, 1),
        "IS_CLKDIVP_INVERTED": random.randint(0, 1),
        "IS_CLKDIV_INVERTED": random.randint(0, 1),
        "IS_CLKB_INVERTED": random.randint(0, 1),
        "IS_CLK_INVERTED": random.randint(0, 1),
        "DYN_CLKDIV_INV_EN": verilog.quote(random.choice(["TRUE", "FALSE"])),
        "DYN_CLK_INV_EN": verilog.quote(random.choice(["TRUE", "FALSE"])),
        "IOBDELAY": verilog.quote(
            random.choice(["NONE", "IBUF", "IFD", "BOTH"])),
        "OFB_USED": verilog.quote(
            random.choice(["TRUE"] + ["FALSE"] * 9)),  # Force more FALSEs
    }

    iface_type = random.choice(
        ["NETWORKING", "OVERSAMPLE", "MEMORY", "MEMORY_DDR3", "MEMORY_QDR"])
    data_rate = random.choice(["SDR", "DDR"])
    serdes_mode = random.choice(["MASTER", "SLAVE"])

    params["INTERFACE_TYPE"] = verilog.quote(iface_type)
    params["DATA_RATE"] = verilog.quote(data_rate)
    params["SERDES_MODE"] = verilog.quote(serdes_mode)

    # Networking mode
    if iface_type == "NETWORKING":
        data_widths = {
            "SDR": [2, 3, 4, 5, 6, 7, 8],
            "DDR": [4, 6, 8, 10, 14],
        }
        params["DATA_WIDTH"] = random.choice(data_widths[data_rate])

    # Others
    else:
        params["DATA_WIDTH"] = 4

    if verilog.unquote(params["OFB_USED"]) == "TRUE":
        params["IOBDELAY"] = verilog.quote("NONE")

    return params


def gen_iddr(loc):

    # Site params
    params = {
        "SITE_LOC":
        verilog.quote(loc),
        "USE_IDELAY":
        random.randint(0, 1),
        "BEL_TYPE":
        verilog.quote(random.choice(["IDDR", "IDDR_NO_CLK"])),
        "INIT_Q1":
        random.randint(0, 1),
        "INIT_Q2":
        random.randint(0, 1),
        "SRTYPE":
        verilog.quote(random.choice(["ASYNC", "SYNC"])),
        "DDR_CLK_EDGE":
        verilog.quote(
            random.choice(
                ["OPPOSITE_EDGE", "SAME_EDGE", "SAME_EDGE_PIPELINED"])),
        "CE1USED":
        random.randint(0, 1),
        "SR_MODE":
        verilog.quote(random.choice(["NONE", "SET", "RST"])),
        "IS_C_INVERTED":
        random.randint(0, 1),
        "IS_D_INVERTED":
        random.randint(0, 1),
    }

    if params["USE_IDELAY"]:
        params["IDELMUX"] = random.randint(0, 1)
        params["IFFDELMUX"] = random.randint(0, 1)
    else:
        params["IDELMUX"] = 0
        params["IFFDELMUX"] = 0

    return params


def run():

    # Get all [LR]IOI3 tiles
    tiles = list(gen_sites())

    # Header
    print("// Tile count: %d" % len(tiles))
    print("// Seed: '%s'" % os.getenv("SEED"))
    print(
        '''
module top (
  (* CLOCK_BUFFER_TYPE = "NONE" *)
  input  wire clk1,
  (* CLOCK_BUFFER_TYPE = "NONE" *)
  input  wire clk2,
  input  wire ce,
  input  wire rst,
  input  wire [{N}:0] di,
  output wire [{N}:0] do
);

wire [{N}:0] di_buf;
wire [{N}:0] do_buf;

// IDELAYCTRL
(* KEEP, DONT_TOUCH *)
IDELAYCTRL idelayctrl();
    '''.format(**{"N": len(tiles) - 1}))

    # LOCes IOBs
    data = []
    for i, sites in enumerate(tiles):
        tile_name = sites[0]

        # Use site
        if random.randint(0, 19) > 0:  # Use more often

            # Top sites
            if random.randint(0, 1):
                this_sites = sites[1]
                other_sites = sites[2]
            # Bottom sites
            else:
                this_sites = sites[2]
                other_sites = sites[1]

            # Generate cell
            bel_types = ["IDDR", "ISERDESE2"]
            bel_type = bel_types[int(
                random.randint(0, 2) > 0)]  # ISERDES more often
            if bel_type == "ISERDESE2":
                params = gen_iserdes(this_sites["ILOGIC"])
            if bel_type == "IDDR":
                params = gen_iddr(this_sites["ILOGIC"])

            params["IDELAY_LOC"] = verilog.quote(this_sites["IDELAY"])
            params["IS_USED"] = 1

            # Instantiate the cell
            print('')
            print('// This : ' + " ".join(this_sites.values()))
            print('// Other: ' + " ".join(other_sites.values()))
            print('(* LOC="%s", KEEP, DONT_TOUCH *)' % this_sites["IOB"])
            print('IBUF ibuf_%03d (.I(di[%3d]), .O(di_buf[%3d]));' % (i, i, i))
            print('(* LOC="%s", KEEP, DONT_TOUCH *)' % other_sites["IOB"])
            print('OBUF obuf_%03d (.I(do_buf[%3d]), .O(do[%3d]));' % (i, i, i))

            clk1_conn = random.choice(["clk1", ""])

            param_str = ",".join(".%s(%s)" % (k, v) for k, v in params.items())
            print(
                'ilogic_single #(%s) ilogic_%03d (.clk1(%s), .clk2(clk2), .ce(ce), .rst(rst), .I(di_buf[%3d]), .O(do_buf[%3d]));'
                % (param_str, i, clk1_conn, i, i))

            params["CHAINED"] = 0
            params["TILE_NAME"] = tile_name

            # Params for the second site
            other_params = {
                "TILE_NAME": tile_name,
                "SITE_LOC": verilog.quote(other_sites["ILOGIC"]),
                "IDELAY_LOC": verilog.quote(other_sites["IDELAY"]),
                "IS_USED": 0,
            }

            # Append to data list
            data.append([params, other_params])

        # Don't use sites
        else:

            params_list = [
                {
                    "TILE_NAME": tile_name,
                    "SITE_LOC": verilog.quote(sites[1]["ILOGIC"]),
                    "IDELAY_LOC": verilog.quote(sites[1]["IDELAY"]),
                    "IS_USED": 0,
                },
                {
                    "TILE_NAME": tile_name,
                    "SITE_LOC": verilog.quote(sites[2]["ILOGIC"]),
                    "IDELAY_LOC": verilog.quote(sites[2]["IDELAY"]),
                    "IS_USED": 0,
                }
            ]

            data.append(params_list)

    # Store params
    with open("params.json", "w") as fp:
        json.dump(data, fp, sort_keys=True, indent=1)

    print(
        '''
endmodule

(* KEEP, DONT_TOUCH *)
module ilogic_single(
  input  wire clk1,
  input  wire clk2,
  input  wire ce,
  input  wire rst,
  input  wire I,
  output wire O,
  input  wire [1:0] shiftin,
  output wire [1:0] shiftout
);

parameter SITE_LOC = "";
parameter IS_USED = 1;
parameter BEL_TYPE = "ISERDESE2";
parameter IDELAY_LOC = "";
parameter USE_IDELAY = 0;
parameter IDELMUX = 0;
parameter IFFDELMUX = 0;
parameter INTERFACE_TYPE = "NETWORKING";
parameter DATA_RATE = "DDR";
parameter DATA_WIDTH = 4;
parameter SERDES_MODE = "MASTER";
parameter NUM_CE = 2;
parameter INIT_Q1 = 0;
parameter INIT_Q2 = 0;
parameter INIT_Q3 = 0;
parameter INIT_Q4 = 0;
parameter SRVAL_Q1 = 0;
parameter SRVAL_Q2 = 0;
parameter SRVAL_Q3 = 0;
parameter SRVAL_Q4 = 0;
parameter IS_D_INVERTED = 0;
parameter IS_OCLK_INVERTED = 0;
parameter IS_OCLKB_INVERTED = 0;
parameter IS_CLK_INVERTED = 0;
parameter IS_CLKB_INVERTED = 0;
parameter IS_CLKDIV_INVERTED = 0;
parameter IS_CLKDIVP_INVERTED = 0;
parameter DYN_CLKDIV_INV_EN = "FALSE";
parameter DYN_CLK_INV_EN = "FALSE";
parameter IOBDELAY = "NONE";
parameter OFB_USED = "FALSE";
parameter DDR_CLK_EDGE = "OPPOSITE_EDGE";
parameter SRTYPE = "ASYNC";
parameter CE1USED = 0;
parameter SR_MODE = "NONE";
parameter IS_C_INVERTED = 0;

wire [8:0] x;
wire ddly;

(* KEEP, DONT_TOUCH *)
generate if (IS_USED && USE_IDELAY) begin

  // IDELAY
  (* LOC=IDELAY_LOC, KEEP, DONT_TOUCH *)
  IDELAYE2 idelay
  (
    .C(clk),
    .REGRST(),
    .LD(),
    .CE(),
    .INC(),
    .CINVCTRL(),
    .CNTVALUEIN(),
    .IDATAIN(I),
    .DATAIN(),
    .LDPIPEEN(),
    .DATAOUT(ddly),
    .CNTVALUEOUT()
  );

end else begin

  assign ddly = 0;

end endgenerate

(* KEEP, DONT_TOUCH *)
generate if (IS_USED && BEL_TYPE == "ISERDESE2") begin

  // ISERDES
  (* LOC=SITE_LOC, KEEP, DONT_TOUCH *)
  ISERDESE2 #
  (
  .INTERFACE_TYPE(INTERFACE_TYPE),
  .DATA_RATE(DATA_RATE),
  .DATA_WIDTH(DATA_WIDTH),
  .SERDES_MODE(SERDES_MODE),
  .NUM_CE(NUM_CE),
  .IS_D_INVERTED(IS_D_INVERTED),
  .IS_OCLK_INVERTED(IS_OCLK_INVERTED),
  .IS_OCLKB_INVERTED(IS_OCLKB_INVERTED),
  .IS_CLK_INVERTED(IS_CLK_INVERTED),
  .IS_CLKB_INVERTED(IS_CLKB_INVERTED),
  .IS_CLKDIV_INVERTED(IS_CLKDIV_INVERTED),
  .IS_CLKDIVP_INVERTED(IS_CLKDIVP_INVERTED),
  .INIT_Q1(INIT_Q1),
  .INIT_Q2(INIT_Q2),
  .INIT_Q3(INIT_Q3),
  .INIT_Q4(INIT_Q4),
  .SRVAL_Q1(SRVAL_Q1),
  .SRVAL_Q2(SRVAL_Q2),
  .SRVAL_Q3(SRVAL_Q3),
  .SRVAL_Q4(SRVAL_Q4),
  .DYN_CLKDIV_INV_EN(DYN_CLKDIV_INV_EN),
  .DYN_CLK_INV_EN(DYN_CLK_INV_EN),
  .IOBDELAY(IOBDELAY),
  .OFB_USED(OFB_USED)
  )
  isedres
  (
  .D(I),
  .DDLY(),
  .OFB(),
  //.TFB(),
  .CE1(),
  .CE2(),
  .DYNCLKSEL(),
  .CLK(clk1),
  .CLKB(clk2),
  .OCLK(),
  .OCLKB(),
  .DYNCLKDIVSEL(),
  .CLKDIV(),
  .CLKDIVP(),
  .RST(),
  .BITSLIP(),
  .O(x[8]),
  .Q1(x[0]),
  .Q2(x[1]),
  .Q3(x[2]),
  .Q4(x[3]),
  .Q5(x[4]),
  .Q6(x[5]),
  .Q7(x[6]),
  .Q8(x[7]),
  .SHIFTIN1(shiftin[0]),
  .SHIFTIN2(shiftin[1]),
  .SHIFTOUT1(shiftout[0]),
  .SHIFTOUT2(shiftout[1])
  );

end else if (IS_USED && BEL_TYPE == "IDDR") begin

  // IDDR
  (* LOC=SITE_LOC, KEEP, DONT_TOUCH *)
  IDDR #
  (
  .IS_C_INVERTED(IS_C_INVERTED),
  .IS_D_INVERTED(IS_D_INVERTED),
  .DDR_CLK_EDGE(DDR_CLK_EDGE),
  .INIT_Q1(INIT_Q1),
  .INIT_Q2(INIT_Q2),
  .SRTYPE(SRTYPE)
  )
  iddr
  (
  .C(clk1),
  .CE( (CE1USED) ? ce : 1'hx ),
  .D( (IFFDELMUX) ? ddly : I  ),
  .S( (SR_MODE == "SET") ? rst : 1'd0 ),
  .R( (SR_MODE == "RST") ? rst : 1'd0 ),
  .Q1(x[0]),
  .Q2(x[1])
  );

  assign x[8] = (IDELMUX) ? ddly : I;
  assign x[7:2] = 0;

end else if (IS_USED && BEL_TYPE == "IDDR_NO_CLK") begin

  // IDDR
  (* LOC=SITE_LOC, KEEP, DONT_TOUCH *)
  IDDR #
  (
  .IS_C_INVERTED(IS_C_INVERTED),
  .IS_D_INVERTED(IS_D_INVERTED),
  .DDR_CLK_EDGE(DDR_CLK_EDGE),
  .INIT_Q1(INIT_Q1),
  .INIT_Q2(INIT_Q2),
  .SRTYPE(SRTYPE)
  )
  iddr
  (
  .C(),
  .CE( (CE1USED) ? ce : 1'hx ),
  .D( (IFFDELMUX) ? ddly : I  ),
  .S( (SR_MODE == "SET") ? rst : 1'd0 ),
  .R( (SR_MODE == "RST") ? rst : 1'd0 ),
  .Q1(x[0]),
  .Q2(x[1])
  );

  assign x[8] = (IDELMUX) ? ddly : I;
  assign x[7:2] = 0;

end else begin

  assign x[0] = I;
  assign x[1] = I;
  assign x[2] = I;
  assign x[3] = I;
  assign x[4] = I;
  assign x[5] = I;
  assign x[6] = I;
  assign x[7] = I;
  assign x[8] = I;

end endgenerate

// Output
assign O = |x;

endmodule

    ''')


run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker
from prjxray import util

segmk = Segmaker("design.bits")
cache = dict()

print("Loading tags")
'''
module,loc,n
clb_NFFMUX_O5,SLICE_X12Y100,3
clb_NFFMUX_AX,SLICE_X13Y100,2
clb_NFFMUX_O6,SLICE_X14Y100,3
'''
f = open('params.csv', 'r')
f.readline()
for l in f:
    module, loc, n = l.split(',')
    n = int(n)
    which = chr(ord('A') + n)
    # clb_NFFMUX_AX => AX
    src = module.replace('clb_NFFMUX_', '')
    '''
    AFFMUX
            30_00   30_01   30_02   30_03
    F78     1       1
    CY      1               1
    O5      1                       1
    AX              1
    XOR                     1
    O6                              1
    '''

    # if location not included in cache yet: start with assuming all four MUXes are unused.
    if loc not in cache:
        cache[loc] = set("ABCD")

    # rewrite name of F78 source net: MUXes A and C have an F7 input, MUX B has an F8 input
    if src == "F78":
        if which in "AC":
            src = "F7"
        elif which == "B":
            src = "F8"
        else:
            assert 0

    # rewrite name of AX source net: It's actually AX, BX, CX, or DX
    if src == "AX":
        src = which + "X"

    # add the 1-tag for this connection
    tag = "%sFFMUX.%s" % (which, src)
    segmk.add_site_tag(loc, tag, 1)

    # remove this MUX from the cache, preventing generation of 0-tags for this MUX
    cache[loc].remove(which)

# create 0-tags for all sources on the remaining (unused) MUXes
for loc, muxes in cache.items():
    for which in muxes:
        for src in "F7 F8 CY O5 AX XOR O6 MC31".split():
            if src == "MC31" and which is not "D": continue
            if src == "F7" and which not in "AC": continue
            if src == "F8" and which not in "B": continue
            if src == "AX": src = which + "X"
            tag = "%sFFMUX.%s" % (which, src)
            segmk.add_site_tag(loc, tag, 0)


def bitfilter(frame_idx, bit_idx):

    # Since the SRL32 is enabled along with DFFMUX.MC31, bits related to
    # SRL32 features are masked out.
    if (frame_idx, bit_idx) in [
        (30, 16),  # ALUT.SRL
        (1, 23),  # WEMUX.CE
    ]:
        return False

    return util.bitfilter_clb_mux(frame_idx, bit_idx)


segmk.compile(bitfilter=bitfilter)
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os, random
random.seed(0)
from prjxray import util
from prjxray import verilog

# INCREMENT is the amount of additional CLBN to be instantiated in the design.
# This makes the fuzzer compilation more robust against failures.
INCREMENT = os.getenv('CLBN', 0)
CLBN = 40 + int(INCREMENT)
print('//Requested CLBs: %s' % str(CLBN))


def gen_slices():
    for _tile_name, site_name, _site_type in util.get_roi().gen_sites(
        ['SLICEL', 'SLICEM']):
        yield site_name


DIN_N = CLBN * 8
DOUT_N = CLBN * 8

verilog.top_harness(DIN_N, DOUT_N)

f = open('params.csv', 'w')
f.write('module,loc,n,def_a\n')
slices = gen_slices()
print(
    'module roi(input clk, input [%d:0] din, output [%d:0] dout);' %
    (DIN_N - 1, DOUT_N - 1))
for i in range(CLBN):
    bel = ''

    module = 'clb_N5FFMUX'
    n = random.randint(0, 3)
    def_a = random.randint(0, 1)
    loc = next(slices)

    print('    %s' % module)
    print('            #(.LOC("%s"), .N(%d), .DEF_A(%d))' % (loc, n, def_a))
    print(
        '            clb_%d (.clk(clk), .din(din[  %d +: 8]), .dout(dout[  %d +: 8]));'
        % (i, 8 * i, 8 * i))

    f.write('%s,%s,%s,%s\n' % (module, loc, n, def_a))
f.close()
print(
    '''endmodule

// ---------------------------------------------------------------------

''')

print(
    '''
module clb_N5FFMUX (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_X22Y100";
    parameter N=-1;
    parameter DEF_A=1;
    wire lutdo, lutco, lutbo, lutao;
    wire lut7bo, lut7ao;
    wire lut8o;

    reg [3:0] ffds;
    wire lutdo5, lutco5, lutbo5, lutao5;
    //wire lutno5 [3:0] = {lutao5, lutbo5, lutco5, lutdo5};
    wire lutno5 [3:0] = {lutdo5, lutco5, lutbo5, lutao5};
    always @(*) begin
        if (DEF_A) begin
            //Default poliarty A
            ffds[3] = lutdo5;
            ffds[2] = lutco5;
            ffds[1] = lutbo5;
            ffds[0] = lutao5;
            ffds[N] = din[6];
        end else begin
            //Default polarity B
            ffds[3] = din[6];
            ffds[2] = din[6];
            ffds[1] = din[6];
            ffds[0] = din[6];
            ffds[N] = lutno5[N];
        end
    end

    (* LOC=LOC, BEL="F8MUX", KEEP, DONT_TOUCH *)
    MUXF8 mux8 (.O(), .I0(lut7bo), .I1(lut7ao), .S(din[6]));
    (* LOC=LOC, BEL="F7BMUX", KEEP, DONT_TOUCH *)
    MUXF7 mux7b (.O(lut7bo), .I0(lutdo), .I1(lutco), .S(din[6]));
    (* LOC=LOC, BEL="F7AMUX", KEEP, DONT_TOUCH *)
    MUXF7 mux7a (.O(lut7ao), .I0(lutbo), .I1(lutao), .S(din[6]));

	(* LOC=LOC, BEL="D6LUT", KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_DEAD_0000_0001)
	) lutd (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(lutdo5),
		.O6(lutdo));
	(* LOC=LOC, BEL="D5FF" *)
	FDPE ffd (
		.C(clk),
		.Q(dout[1]),
		.CE(din[0]),
		.PRE(din[1]),
		.D(ffds[3]));

	(* LOC=LOC, BEL="C6LUT", KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_BEEF_0000_0001)
	) lutc (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(lutco5),
		.O6(lutco));
	(* LOC=LOC, BEL="C5FF" *)
	FDPE ffc (
		.C(clk),
		.Q(dout[2]),
		.CE(din[0]),
		.PRE(din[1]),
		.D(ffds[2]));

	(* LOC=LOC, BEL="B6LUT", KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_CAFE_0000_0001)
	) lutb (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(lutbo5),
		.O6(lutbo));
	(* LOC=LOC, BEL="B5FF" *)
	FDPE ffb (
		.C(clk),
		.Q(dout[3]),
		.CE(din[0]),
		.PRE(din[1]),
		.D(ffds[1]));

	(* LOC=LOC, BEL="A6LUT", KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_1CE0_0000_0001)
	) luta (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(lutao5),
		.O6(lutao));
	(* LOC=LOC, BEL="A5FF" *)
	FDPE ffa (
		.C(clk),
		.Q(dout[4]),
		.CE(din[0]),
		.PRE(din[1]),
		//D can only come from O5 or AX
		//AX is used by MUXF7:S
		.D(ffds[0]));
endmodule
''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Load tileconn.json and node_wires.json and verify both node names and
node <-> wire mapping against raw node data. """

import argparse
import datetime
import json
import multiprocessing
import os.path
import progressbar
import pyjson5 as json5

from prjxray import util, lib
from prjxray.grid import Grid
from prjxray.connections import Connections
from prjxray.node_model import NodeModel


def read_json5(fname):
    with open(fname, 'r') as f:
        return json5.load(f)


def read_raw_node_data(pool, root_dir):
    """ Read raw node data from root dir. """
    _, nodes = lib.read_root_csv(root_dir)

    raw_node_data = []
    with progressbar.ProgressBar(max_value=len(nodes)) as bar:
        for idx, node in enumerate(pool.imap_unordered(
                read_json5,
                nodes,
                chunksize=20,
        )):
            bar.update(idx)
            raw_node_data.append(node)
            bar.update(idx + 1)

    return raw_node_data


def main():
    parser = argparse.ArgumentParser(
        description="Verify tileconn and node_wires.")
    parser.add_argument('--root_dir', required=True)
    parser.add_argument('--output_dir', required=True)
    parser.add_argument('--max_cpu', type=int, default=10)
    parser.add_argument('--ignored_wires')

    args = parser.parse_args()

    print('{} Reading tilegrid'.format(datetime.datetime.now()))
    with open(os.path.join(util.get_db_root(), util.get_fabric(),
                           'tilegrid.json')) as f:
        tilegrid = json.load(f)
        grid = Grid(db=None, tilegrid=tilegrid)

    print('{} Reading tileconn'.format(datetime.datetime.now()))
    with open(os.path.join(args.output_dir, 'tileconn.json')) as f:
        tileconn = json.load(f)

    print(
        '{} Reading tile wires from tile types'.format(
            datetime.datetime.now()))
    tile_wires = {}
    for f in os.listdir(args.output_dir):
        if f.endswith('.json') and f.startswith('tile_type_'):
            if '_site_type_' in f:
                continue

            tile_type = f[len('tile_type_'):-len('.json')]
            with open(os.path.join(args.output_dir, f)) as fin:
                tile_wires[tile_type] = json.load(fin)['wires']

    connections = Connections(
        tilegrid=tilegrid,
        tileconn=tileconn,
        tile_wires=tile_wires,
    )

    print('{} Reading node wires'.format(datetime.datetime.now()))
    with open(os.path.join(args.output_dir, 'node_wires.json')) as f:
        node_wires = json.load(f)

    print('{} Build initial node model'.format(datetime.datetime.now()))
    node_model = NodeModel(
        grid=grid,
        connections=connections,
        tile_wires=tile_wires,
        node_wires=node_wires,
        progressbar=progressbar.progressbar)

    print('{} Build node model'.format(datetime.datetime.now()))
    nodes = set(node_model.get_nodes())

    print('{} Read raw node data for testing'.format(datetime.datetime.now()))

    processes = min(multiprocessing.cpu_count(), args.max_cpu)
    with multiprocessing.Pool(processes=processes) as pool:
        raw_node_data = read_raw_node_data(pool, args.root_dir)

    print('{} Read ignored wires list'.format(datetime.datetime.now()))
    ignored_wires = []
    ignored_wires_file = args.ignored_wires
    if os.path.exists(ignored_wires_file):
        with open(ignored_wires_file) as f:
            ignored_wires = set(tuple(l.strip().split('/')) for l in f)

    print(
        '{} Verify nodes against raw node data'.format(
            datetime.datetime.now()))
    for node in progressbar.progressbar(raw_node_data):
        tile, wire = node['node'].split('/')

        assert (tile, wire) in nodes
        wires_for_model = node_model.get_wires_for_node(tile, wire)

        wires = set()
        for wire in node['wires']:
            wire_tile, wire_name = wire['wire'].split('/')
            wires.add((wire_tile, wire_name))

        if len(wires) != len(wires_for_model):
            wires2 = set(wires_for_model)
            a_minus_b = wires - wires2
            b_minus_a = wires2 - wires

            assert len(b_minus_a) == 0

            assert len(a_minus_b - ignored_wires) == 0, a_minus_b

        for tile, wire in wires_for_model:
            assert (tile, wire) in wires


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json

with open('output/error_nodes.json') as f:
    flat_error_nodes = json.load(f)

error_nodes = {}
for node, raw_node, generated_nodes in flat_error_nodes:
    if node not in error_nodes:
        error_nodes[node] = {
            'raw_node': set(raw_node),
            'generated_nodes': set(),
        }

    assert error_nodes[node]['raw_node'] == set(raw_node)
    error_nodes[node]['generated_nodes'].add(tuple(sorted(generated_nodes)))

for node, error in error_nodes.items():
    combined_generated_nodes = set()
    for generated_node in error['generated_nodes']:
        combined_generated_nodes |= set(generated_node)

    assert error['raw_node'] == combined_generated_nodes, (node, error)

    good_node = max(error['generated_nodes'], key=lambda x: len(x))
    bad_nodes = error['generated_nodes'] - set((good_node, ))

    if max(len(generated_node) for generated_node in bad_nodes) > 1:
        assert False, node
    else:
        for generated_node in bad_nodes:
            for wire in generated_node:
                print(wire)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")

print("Loading tags")
'''
module,loc,n,def_a
clb_N5FFMUX,SLICE_X12Y100,3,1
clb_N5FFMUX,SLICE_X13Y100,0,1
clb_N5FFMUX,SLICE_X14Y100,3,1
'''
f = open('params.csv', 'r')
f.readline()
for l in f:
    module, loc, n, def_a = l.split(',')
    def_a = int(def_a)
    n = int(n)
    #which = chr(ord('A') + n)

    for i, which in enumerate('ABCD'):
        # Theory: there is one bit for each mux positon
        # In each config 3 muxes are in one position, other 3 are in another
        inv = int(i == n)
        segmk.add_site_tag(loc, "%c5FFMUX.IN_A" % which, def_a ^ inv)
        segmk.add_site_tag(loc, "%c5FFMUX.IN_B" % which, 1 ^ def_a ^ inv)
segmk.compile()
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Database files available for a tile type. """
from collections import namedtuple
import json
from prjxray import lib
from prjxray.timing import fast_slow_tuple_to_corners, RcElement
from prjxray.util import OpenSafeFile

TileDbs = namedtuple(
    'TileDbs', 'segbits block_ram_segbits ppips mask tile_type')


class OutPinTiming(namedtuple('OutPinTiming', 'delays drive_resistance')):
    """ Timing for site output pins.

    Attributes
    ----------
    delays : dicts of PvtCorner to IntristicDelay
        Intristic delay of output pin.
    drive_resistance : float
        Resistance of drive output pin (milliOhms).

    """
    pass


class InPinTiming(namedtuple('InPinTiming', 'delays capacitance')):
    """ Timing for site input pins.

    Attributes
    ----------
    delays : dicts of PvtCorner to IntristicDelay
        Intristic delay of input pin.
    capacitance : float
        Capacitance of input pints (microFarads).

    """
    pass


class PipTiming(namedtuple('PipTiming',
                           'delays drive_resistance internal_capacitance')):
    """ Timing for pips.

    Attributes
    ----------
    delays : dicts of PvtCorner to IntristicDelay
        Intristic delay of pip.
    internal_capacitance : float
        Capacitance (microFarads) of pip (which is only seen if pip is used).
    drive_resistance : float
        Resistance of drive output pin (milliOhms).

    """
    pass


class Pip(namedtuple(
        'Pip',
    ('name', 'net_to', 'net_from', 'can_invert', 'is_directional', 'is_pseudo',
     'is_pass_transistor', 'timing', 'backward_timing'))):
    """ Pip information.

    Attributes
    ----------

    name : str
        Name of pip
    net_to : str
        Name of output tile wire when pip is unidirectional.
    net_from: str
        Name of input tile wire when pip is unidirectional.
    can_invert : bool
        Can this pip invert the signal.
    is_directional : bool
        True if this pip is unidirectional, False if this pip is
        unidirectional.
    is_pseudo : bool
        True if this pip is mark as a pseudo-pip.
    is_pass_transistor : bool
        True if this pip is non-isolating.
    timing : PipTiming
        Timing used when connecting net_from to net_to.  This is the only
        timing used when a pip is unidirectional.

        May be None if timing information is not present in the database.
    backward_timing : PipTiming
        Timing used when connecting net_to to net_from.  This is only used
        if the pip is bidirectional.

        May be None if timing information is not present in the database.

    """
    pass


class Site(namedtuple('Site', 'name prefix x y type site_pins')):
    """ Represents an instance of a site within a tile.

    Attributes
    ----------
    name : str
        Name of site within tile, instance specific.
    prefix : str
        Prefix of site naming in Xilinx parlance.
    type : str
        What type of slice this instance presents.
    site_pins : list of SitePin
        Instaces of site pins within this site and tile.  This is an tuple of
        SitePin tuples, and is specific to this instance of the site within
        the tile.

    """


class SitePin(namedtuple('SitePin', 'name wire timing')):
    """ Tuple representing a site pin within a tile.

    Sites are generic based on type, however sites are instanced
    within a tile 1 or more times.  The SitePin contains both site type generic
    information and tile type specific information.

    Attributes
    ----------
    name : str
        Site type specific name.  This name is expected to be the same for
        all sites of the same type.
    wire : str
        Wire name within the tile.  This name is site instance specific.
    timing : Either InPinTiming or OutPinTiming
        Timing of site pin. May be None if database lacks timing information.

    """


WireInfo = namedtuple('WireInfo', 'pips sites')

# Conversion factor from database to internal units.
RESISTANCE_FACTOR = 1e3
CAPACITANCE_FACTOR = 1e3


def get_pip_timing(pip_timing_json):
    """ Convert pip_timing_json JSON into PipTiming object.

    Returns
    -------
    If timing information is not present for this pip, returns None.
    If timing information is present, returns PipTiming.  Some fields may be
    None if the pip type lacks that field.

    """

    if pip_timing_json is None:
        return None

    delays = None

    if pip_timing_json.get('delay') is not None:
        delays = fast_slow_tuple_to_corners(pip_timing_json.get('delay'))

    in_cap = pip_timing_json.get('in_cap')
    if in_cap is not None:
        in_cap = float(in_cap) / CAPACITANCE_FACTOR
    else:
        in_cap = 0

    res = pip_timing_json.get('res')
    if res is not None:
        res = float(res) / RESISTANCE_FACTOR
    else:
        res = 0

    return PipTiming(
        delays=delays,
        drive_resistance=res,
        internal_capacitance=in_cap,
    )


def get_site_pin_timing(site_pin_info):
    """ Convert site_pin_info JSON into InPinTiming or OutPinTiming object.

    Returns
    -------
    If timing information is not present for this site pin, returns None.
    If this is an output pin, returns OutPinTiming.
    If this is an input pin, returns InPinTiming.

    """
    if isinstance(site_pin_info, str):
        return site_pin_info, None

    wire = site_pin_info['wire']

    if 'delay' not in site_pin_info:
        return None

    delays = fast_slow_tuple_to_corners(site_pin_info['delay'])

    if 'cap' in site_pin_info:
        assert 'res' not in site_pin_info
        return wire, InPinTiming(
            delays=delays,
            capacitance=float(site_pin_info['cap']) / CAPACITANCE_FACTOR,
        )
    else:
        assert 'res' in site_pin_info
        return wire, OutPinTiming(
            delays=delays,
            drive_resistance=float(site_pin_info['res']) / RESISTANCE_FACTOR,
        )


def get_wires(wires):
    """ Converts database input to dictionary of tile wires to wire timing.

    Returns dictionary of tile wire name to RcElement or None. """

    if isinstance(wires, list):
        # Handle old database gracefully.
        return {wire: None for wire in wires}

    output = {}

    for wire, rc_json in wires.items():
        if rc_json is None:
            output[wire] = RcElement(
                resistance=0,
                capacitance=0,
            )
        else:
            output[wire] = RcElement(
                resistance=float(rc_json['res']) / RESISTANCE_FACTOR,
                capacitance=float(rc_json['cap']) / CAPACITANCE_FACTOR,
            )

    return output


def is_pass_transistor(pip_json):
    """ Returns boolean if pip JSON indicates pip is a pass transistor.

    Always returns False if database lacks this information.
    """
    if 'is_pass_transistor' in pip_json:
        return bool(int(pip_json['is_pass_transistor']))
    else:
        return False


class Tile(object):
    """ Provides abstration of a tile in the database. """

    def __init__(self, tilename, tile_dbs):
        self.tilename = tilename
        self.tilename_upper = self.tilename.upper()
        self.tile_dbs = tile_dbs

        self.wires = None
        self.sites = None
        self.pips = None
        self.pips_by_name = {}

        def yield_sites(sites):
            for site in sites:
                site_pins = []
                for name, site_pin_info in site['site_pins'].items():
                    if site_pin_info is not None:
                        wire, timing = get_site_pin_timing(site_pin_info)
                        site_pins.append(
                            SitePin(
                                name=name,
                                wire=wire,
                                timing=timing,
                            ))
                    else:
                        site_pins.append(
                            SitePin(
                                name=name,
                                wire=None,
                                timing=None,
                            ))

                yield Site(
                    name=site['name'],
                    prefix=site['prefix'],
                    type=site['type'],
                    x=site['x_coord'],
                    y=site['y_coord'],
                    site_pins=site_pins,
                )

        def yield_pips(pips):
            for name, pip in pips.items():
                yield Pip(
                    name=name,
                    net_to=pip['dst_wire'],
                    net_from=pip['src_wire'],
                    can_invert=bool(int(pip['can_invert'])),
                    is_directional=bool(int(pip['is_directional'])),
                    is_pseudo=bool(int(pip['is_pseudo'])),
                    is_pass_transistor=is_pass_transistor(pip),
                    timing=get_pip_timing(pip.get('src_to_dst')),
                    backward_timing=get_pip_timing(pip.get('dst_to_src')),
                )

        with OpenSafeFile(self.tile_dbs.tile_type) as f:
            tile_type = json.load(f)
            assert self.tilename_upper == tile_type['tile_type']
            self.wires = get_wires(tile_type['wires'])
            self.sites = tuple(yield_sites(tile_type['sites']))
            self.pips = tuple(yield_pips(tile_type['pips']))

        self.wire_info = {}

    def get_wires(self):
        """Returns a set of wire names present in this tile."""
        return self.wires

    def get_sites(self):
        """ Returns tuple of Site namedtuple's present in this tile. """
        return self.sites

    def get_pips(self):
        """ Returns tuple of Pip namedtuple's representing the PIPs in this tile.
    """
        return self.pips

    def get_pip_by_name(self, name):
        if len(self.pips_by_name) == 0:
            for pip in self.pips:
                self.pips_by_name[pip.name] = pip

        return self.pips_by_name[name]

    def get_wire_info(self, target_wire, allow_pseudo=False):
        if len(self.wire_info) == 0:
            for wire in self.wires:
                pips = list()
                sites = list()

                for site in self.sites:
                    for site_pin in site.site_pins:
                        if site_pin.wire == wire:
                            sites.append((site.name, site_pin.name))

                for pip in self.pips:
                    pseudo_filter = (not pip.is_pseudo) or allow_pseudo
                    if (wire == pip.net_to
                            or wire == pip.net_from) and pseudo_filter:
                        pips.append(pip.name)

                assert wire not in self.wire_info
                self.wire_info[wire] = WireInfo(pips=pips, sites=sites)

        return self.wire_info[target_wire]

    def get_instance_sites(self, grid_info):
        """ get_sites returns abstract sites for all tiles of type.
            get_instance_sites converts site info from generic to specific
            based on a tile location.
            """

        site_names = set()

        for site in self.sites:
            site_name = '{}_X{}Y{}'.format(site.prefix, site.x, site.y)
            origin_x, origin_y = lib.find_origin_coordinate(
                site_name, grid_info.sites.keys())

            x = site.x + origin_x
            y = site.y + origin_y

            site_name = '{}_X{}Y{}'.format(site.prefix, x, y)

            if site_name not in grid_info.sites:
                type_count = 0
                for site_name_from_grid, site_type in grid_info.sites.items():
                    if site.type == site_type:
                        type_count += 1
                        site_name = site_name_from_grid

                assert type_count == 1, (site_name, type_count)

            site_names.add(site_name)
            assert site.type == grid_info.sites[site_name]

            yield Site(
                name=site_name,
                prefix=site.prefix,
                type=site.type,
                x=x,
                y=y,
                site_pins=site.site_pins,
            )

        assert site_names == set(grid_info.sites.keys())


def get_other_wire_from_pip(pip, wire):
    if wire == pip.net_to:
        return pip.net_from
    elif wire == pip.net_from:
        return pip.net_to
    else:
        assert False, (pip, wire)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import random
random.seed(0)
import os
import re
from prjxray import util
from prjxray import verilog

# INCREMENT is the amount of additional CLBN to be instantiated in the design.
# This makes the fuzzer compilation more robust against failures.
INCREMENT = os.getenv('CLBN', 0)
CLBN = 600 + int(INCREMENT)
print('//Requested CLBs: %s' % str(CLBN))


def gen_slices():
    for _tile_name, site_name, _site_type in util.get_roi().gen_sites(
        ['SLICEL', 'SLICEM']):
        yield site_name


DIN_N = CLBN * 4
DOUT_N = CLBN * 1
ffprims = ('FDRE', )
ff_bels = (
    'AFF',
    'A5FF',
    'BFF',
    'B5FF',
    'CFF',
    'C5FF',
    'DFF',
    'D5FF',
)

verilog.top_harness(DIN_N, DOUT_N)

f = open('params.csv', 'w')
f.write('name,loc,ce,r\n')
slices = gen_slices()
print(
    'module roi(input clk, input [%d:0] din, output [%d:0] dout);' %
    (DIN_N - 1, DOUT_N - 1))
for i in range(CLBN):
    ffprim = random.choice(ffprims)
    force_ce = random.randint(0, 1)
    force_r = random.randint(0, 1)
    # clb_FD clb_FD (.clk(clk), .din(din[  0 +: 4]), .dout(dout[  0]));
    # clb_FD_1 clb_FD_1 (.clk(clk), .din(din[  4 +: 4]), .dout(dout[  1]));
    loc = next(slices)
    #bel = random.choice(ff_bels)
    bel = "AFF"
    name = 'clb_%s' % ffprim
    print('    %s' % name)
    print(
        '            #(.LOC("%s"), .BEL("%s"), .FORCE_CE1(%d), .nFORCE_R0(%d))'
        % (loc, bel, force_ce, force_r))
    print(
        '            clb_%d (.clk(clk), .din(din[  %d +: 4]), .dout(dout[  %d]));'
        % (i, 4 * i, 1 * i))
    f.write('%s,%s,%s,%s\n' % (name, loc, force_ce, force_r))
f.close()
print(
    '''endmodule

// ---------------------------------------------------------------------

''')

print(
    '''
module clb_FDRE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y114";
    parameter BEL="AFF";
    parameter FORCE_CE1=0;
    parameter nFORCE_R0=1;
    (* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH *)
    FDRE ff (
        .C(clk),
        .Q(dout),
        .CE(din[0] | FORCE_CE1),
        .R(din[1] & nFORCE_R0),
        .D(din[2])
    );
    endmodule
''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
import argparse

from prjxray.util import OpenSafeFile

def get_elems_count(timings, slice, site, bel_type):
    combinational = 0
    sequential = 0
    for delay in timings[slice][site][bel_type]:
        if 'sequential' in timings[slice][site][bel_type][delay]:
            sequential += 1
        else:
            combinational += 1
    return combinational, sequential


def produce_sdf(timings, outdir):

    for slice in timings:
        sdf = \
"""
(DELAYFILE
    (SDFVERSION \"3.0\")
    (TIMESCALE 1ns)
"""
        for site in sorted(timings[slice]):
            for bel_type in sorted(timings[slice][site]):
                combinational, sequential = get_elems_count(
                    timings, slice, site, bel_type)
                #define CELL
                cell= \
"""
    (CELL
        (CELLTYPE \"{name}\")
        (INSTANCE {location})""".format(name=bel_type.upper(), location=site)
                sdf += cell

                #define delay header (if needed)
                if combinational > 0:
                    delay_hdr = \
"""
        (DELAY
            (ABSOLUTE"""
                    sdf += delay_hdr
                    # add all delays definitions
                    for delay in sorted(timings[slice][site][bel_type]):
                        if 'sequential' in timings[slice][site][bel_type][
                                delay]:
                            continue
                        dly = \
"""
                (IOPATH {input} {output} ({FAST_MIN}::{FAST_MAX})({SLOW_MIN}::{SLOW_MAX}))""".format(**timings[slice][site][bel_type][delay])
                        if 'extra_ports' in timings[slice][site][bel_type][
                                delay] is not None:
                            dly += \
""" #extra ports {}""".format(timings[slice][site][bel_type][delay]['extra_ports'])

                        sdf += dly

                    # close DELAY definition
                    enddelay = \
"""
            )
        )"""
                    sdf += enddelay

                # define TIMINGCHECK header (if needed)
                if sequential > 0:
                    timingcheck_hdr = \
"""
        (TIMINGCHECK"""
                    sdf += timingcheck_hdr

                    for delay in sorted(timings[slice][site][bel_type]):
                        if 'sequential' not in timings[slice][site][bel_type][
                                delay]:
                            continue
                        timingcheck = \
"""
            ({prop} {input} (posedge {clock}) ({SLOW_MIN}::{SLOW_MAX}))""".format(
                        prop=timings[slice][site][bel_type][delay]['sequential'].upper(),
                        **timings[slice][site][bel_type][delay])

                        if 'extra_ports' in timings[slice][site][bel_type][
                                delay] is not None:
                            timingcheck += \
""" #extra ports {}""".format(timings[slice][site][bel_type][delay]['extra_ports'])

                        sdf += timingcheck

                    # close TIMINGCHECK definition
                    endtimingcheck = \
"""
        )"""
                    sdf += endtimingcheck

                endcell = \
"""
    )"""
                sdf += endcell
        # end of SDF
        sdf += \
"""
)"""

        with OpenSafeFile(outdir + '/' + slice + '.sdf', "w") as fp:
            fp.write(sdf)


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument('--json', type=str, help="Input JSON file")
    parser.add_argument('--sdf', type=str, help="SDF files output directory")

    args = parser.parse_args()

    with OpenSafeFile(args.json, 'r') as fp:
        timings = json.load(fp)

    produce_sdf(timings, args.sdf)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def read_site_to_cmt():
    """ Yields clock sources and which CMT they route within. """
    fuzdir = os.getenv('FUZDIR')
    part = os.getenv('XRAY_PART')

    with open(os.path.join(fuzdir, 'build_{}'.format(part),
                           'cmt_regions.csv')) as f:
        for l in f:
            site, cmt = l.strip().split(',')
            yield (site, cmt)


class ClockSources(object):
    """ Class for tracking clock sources.
    """

    def __init__(self, limit=14):
        self.sources = {}
        self.source_to_cmt = {}
        self.used_sources_from_cmt = {}
        self.limit = limit

    def add_clock_source(self, source, cmt):
        """ Adds a source from a specific CMT.

        """
        if cmt not in self.sources:
            self.sources[cmt] = []

        self.sources[cmt].append(source)
        self.source_to_cmt[source] = cmt

    def sources_depleted(self, cmt):
        if cmt in self.sources:
            if cmt not in self.used_sources_from_cmt:
                return False

            return self.sources[cmt] == self.used_sources_from_cmt[cmt]

        return True

    def get_random_source(self, cmt, no_repeats=True):
        """ Get a random source that is routable to the specific CMT.

        get_random_source will return a source that is either cmt='ANY',
        cmt equal to the input CMT, or the adjecent CMT.

        """

        choices = []

        if cmt in self.sources:
            choices.extend(self.sources[cmt])

        random.shuffle(choices)
        for source in choices:

            source_cmt = self.source_to_cmt[source]

            if source_cmt not in self.used_sources_from_cmt:
                self.used_sources_from_cmt[source_cmt] = set()

            if no_repeats and source in self.used_sources_from_cmt[source_cmt]:
                continue

            if len(self.used_sources_from_cmt[source_cmt]) >= self.limit:
                continue

            self.used_sources_from_cmt[source_cmt].add(source)
            return source

        return None


def print_bufhce(name, net):
    print(
        """
(* KEEP, DONT_TOUCH, LOC="{site}" *)
BUFHCE {site} (
.I({clock})
);""".format(site=name, clock=net))


def main():
    """
    GTP_COMMON_MID has clock pips from:

    14 clocks lines from the HROW spine
    """

    cmt_clock_sources = ClockSources()
    site_to_cmt = dict(read_site_to_cmt())
    clock_region_limit = dict()
    clock_region_serdes_location = dict()

    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    def gen_sites(desired_site_type):
        for tile_name in sorted(grid.tiles()):
            loc = grid.loc_of_tilename(tile_name)
            gridinfo = grid.gridinfo_at_loc(loc)
            for site, site_type in gridinfo.sites.items():
                if site_type == desired_site_type:
                    yield tile_name, site

    clock_region_sites = set()

    def get_clock_region_site(site_type, clk_reg):
        for site_name, reg in site_to_cmt.items():
            if site_name.startswith(site_type) and reg in clk_reg:
                if site_name not in clock_region_sites:
                    clock_region_sites.add(site_name)
                    return site_name

    cmt_with_gtp = set()
    for tile_name, site in gen_sites('GTPE2_COMMON'):
        cmt_with_gtp.add(site_to_cmt[site])

    print('''module top();

(* KEEP, DONT_TOUCH *)
LUT6 dummy();
''')

    for _, site in gen_sites('MMCME2_ADV'):
        if site_to_cmt[site] not in cmt_with_gtp:
            continue

        mmcm_clocks = [
            'mmcm_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(7)
        ]

        for clk in mmcm_clocks:
            cmt_clock_sources.add_clock_source(clk, site_to_cmt[site])

        print(
            """
    wire cin1_{site}, cin2_{site}, {c0}, {c1}, {c2}, {c3}, {c4}, {c5};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    MMCME2_ADV pll_{site} (
        .CLKIN1(cin1_{site}),
        .CLKIN2(cin2_{site}),
        .CLKOUT0({c0}),
        .CLKOUT1({c1}),
        .CLKOUT2({c2}),
        .CLKOUT3({c3}),
        .CLKOUT4({c4}),
        .CLKOUT5({c5}),
        .CLKOUT6({c6})
    );""".format(
                site=site,
                c0=mmcm_clocks[0],
                c1=mmcm_clocks[1],
                c2=mmcm_clocks[2],
                c3=mmcm_clocks[3],
                c4=mmcm_clocks[4],
                c5=mmcm_clocks[5],
                c6=mmcm_clocks[6]))

    for _, site in gen_sites('PLLE2_ADV'):
        if site_to_cmt[site] not in cmt_with_gtp:
            continue

        pll_clocks = [
            'pll_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(7)
        ]

        for clk in pll_clocks:
            cmt_clock_sources.add_clock_source(clk, site_to_cmt[site])

        print(
            """
    wire cin1_{site}, cin2_{site}, clkfbin_{site}, {c0}, {c1}, {c2}, {c3}, {c4}, {c5}, {c6};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    PLLE2_ADV pll_{site} (
        .CLKIN1(cin1_{site}),
        .CLKIN2(cin2_{site}),
        .CLKFBIN(clkfbin_{site}),
        .CLKOUT0({c0}),
        .CLKOUT1({c1}),
        .CLKOUT2({c2}),
        .CLKOUT3({c3}),
        .CLKOUT4({c4}),
        .CLKOUT5({c5}),
        .CLKFBOUT({c6})
    );""".format(
                site=site,
                c0=pll_clocks[0],
                c1=pll_clocks[1],
                c2=pll_clocks[2],
                c3=pll_clocks[3],
                c4=pll_clocks[4],
                c5=pll_clocks[5],
                c6=pll_clocks[6],
            ))

    for cmt in cmt_with_gtp:
        for _, bufhce in gen_sites('BUFHCE'):
            if site_to_cmt[bufhce] != cmt:
                continue

            if random.random() < 0.7:
                clock_name = cmt_clock_sources.get_random_source(cmt)
            else:
                continue

            if clock_name is None:
                continue

            print_bufhce("{}".format(bufhce), clock_name)

    print('endmodule')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import sys
import random
import math
from prjxray import util
from prjxray.lut_maker import LutMaker
from prjxray.db import Database
random.seed(int(os.getenv("SEED"), 16))


def bram_count():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    count = 0
    for tile_name in grid.tiles():
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['RAMBFIFO36E1']:
                count += 1

    return count


def sdp_bram(name, width, address_bits):
    depth = 2**address_bits

    return '''
module {name}(
    // Write port
    input wrclk,
    input [{width}-1:0] di,
    input wren,
    input [{address_bits}-1:0] wraddr,
    // Read port
    input rdclk,
    input rden,
    input [{address_bits}-1:0] rdaddr,
    output reg [{width}-1:0] do);

    (* ram_style = "block" *) reg [{width}-1:0] ram[0:{depth}];

    always @ (posedge wrclk) begin
        if(wren == 1) begin
            ram[wraddr] <= di;
        end
    end

    always @ (posedge rdclk) begin
        if(rden == 1) begin
            do <= ram[rdaddr];
        end
    end

endmodule
    '''.format(
        name=name,
        width=width,
        address_bits=address_bits,
        depth=depth,
    )


MAX_BRAM = 8


def emit_sdp_bram(luts, name, modules, lines, width, address_bits):
    modules.append(sdp_bram(name=name, width=width, address_bits=address_bits))

    lines.append(
        '''
    wire [{address_bits}-1:0] {name}_wraddr;
    wire [{address_bits}-1:0] {name}_rdaddr;
    '''.format(
            name=name,
            address_bits=address_bits,
        ))

    for bit in range(address_bits):
        lines.append(
            """
    assign {name}_wraddr[{bit}] = {net};""".format(
                name=name, bit=bit, net=luts.get_next_output_net()))

    for bit in range(address_bits):
        lines.append(
            """
    assign {name}_rdaddr[{bit}] = {net};""".format(
                name=name, bit=bit, net=luts.get_next_output_net()))

    lines.append(
        '''
    (* KEEP, DONT_TOUCH *)
    {name} {name}_inst(
        .wraddr({name}_wraddr),
        .rdaddr({name}_rdaddr)
    );
    '''.format(name=name))

    return width, address_bits, math.ceil(
        float(width) / 72) * 72 * (2**address_bits)


def max_address_bits(width):
    return math.floor(math.log(float(MAX_BRAM * 36 * 1024) / width, 2))


def random_sdp_bram(luts, name, modules, lines):
    sdp_choices = set()

    for width in (1, 2, 4, 8, 16, 18, 32, 36):
        sdp_choices.add((width, (1, max_address_bits(width))))

    for nbram in range(2, MAX_BRAM + 1):
        sdp_choices.add((nbram * 32, (1, max_address_bits(nbram * 32))))
        sdp_choices.add((nbram * 36, (1, max_address_bits(nbram * 36))))
        sdp_choices.add((nbram * 16, (1, max_address_bits(nbram * 16))))
        sdp_choices.add((nbram * 32, (1, max_address_bits(nbram * 32))))

        # Bias some wide but shallow BRAMs to toggle the lower address bits
        # more.
        for address_bits in range(1, 4):
            sdp_choices.add((nbram * 16, (address_bits, address_bits)))

    sdp_choices = sorted(sdp_choices)

    width, address_bits_range = random.choice(sdp_choices)
    address_bits = random.randint(*address_bits_range)
    return emit_sdp_bram(luts, name, modules, lines, width, address_bits)


def run():
    luts = LutMaker()
    count = bram_count()

    max_bram_count = random.randint(1, 200)

    modules = []
    lines = []

    idx = 0
    while count > MAX_BRAM:
        width, address_bits, bits = random_sdp_bram(
            luts, "ram{}".format(idx), modules, lines)

        brams = math.ceil(bits / float(36 * 1024))

        count -= brams

        print(width, address_bits, bits, brams, count, file=sys.stderr)
        idx += 1

        if idx >= max_bram_count:
            break

    for module in modules:
        print(module)

    print('''
module top();
''')

    for lut in luts.create_wires_and_luts():
        print(lut)

    for l in lines:
        print(l)

    print("endmodule")


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import re
import fasm
from prjxray import bitstream


def mk_fasm(tile_name, feature):
    """ Convert matches tile and feature to FasmLine tuple. """
    # Seperate addressing of multi-bit features:
    # TILE.ALUT[0] -> ('TILE', 'ALUT', '0')
    # TILE.ALUT.SMALL -> ('TILE', 'ALUT.SMALL', None)
    m = re.match(r'([A-Za-z0-9_]+).([^\[]+)(\[[0-9]+\])?', feature)
    tag_post = m.group(2)
    address = None
    if m.group(3) is not None:
        address = int(m.group(3)[1:-1])

    feature = '{}.{}'.format(tile_name, tag_post)

    return fasm.FasmLine(
        set_feature=fasm.SetFasmFeature(
            feature=feature,
            start=address,
            end=None,
            value=1,
            value_format=None,
        ),
        annotations=None,
        comment=None)


class FasmDisassembler(object):
    """ Given a Project X-ray data, outputs FasmLine tuples for bits set. """

    def __init__(self, db):
        self.db = db
        self.grid = self.db.grid()
        self.segment_map = self.grid.get_segment_map()
        self.decode_warnings = set()

    def find_features_in_tile(
            self,
            tile_name,
            block_type,
            bits,
            solved_bitdata,
            bitdata,
            verbose=False):
        gridinfo = self.grid.gridinfo_at_tilename(tile_name)

        try:
            tile_segbits = self.grid.get_tile_segbits_at_tilename(tile_name)
        except KeyError as e:
            if not verbose:
                return

            if gridinfo.tile_type in self.decode_warnings:
                return

            comment = " WARNING: failed to load DB for tile type {}".format(
                gridinfo.tile_type)
            yield fasm.FasmLine(
                set_feature=None,
                annotations=None,
                comment=comment,
            )
            yield fasm.FasmLine(
                set_feature=None,
                annotations=(
                    fasm.Annotation('missing_segbits', gridinfo.tile_type),
                    fasm.Annotation('exception', str(e)),
                ),
                comment=None,
            )

            self.decode_warnings.add(gridinfo.tile_type)
            return

        for ones_matched, feature in tile_segbits.match_bitdata(block_type,
                                                                bits, bitdata):
            for frame, bit in ones_matched:
                if frame not in solved_bitdata:
                    solved_bitdata[frame] = set()
                solved_bitdata[frame].add(bit)

            yield mk_fasm(tile_name=tile_name, feature=feature)

    def find_features_in_bitstream(self, bitdata, verbose=False):
        solved_bitdata = {}
        frames = set(bitdata.keys())
        tiles_checked = set()

        emitted_features = set()

        while len(frames) > 0:
            frame = frames.pop()

            # Skip frames that were emptied in a previous iteration.
            if not bitdata[frame]:
                continue

            # Iterate over all tiles that use this frame.
            for bits_info in self.segment_map.segment_info_for_frame(frame):
                # Don't examine a tile twice
                if (bits_info.tile, bits_info.block_type) in tiles_checked:
                    continue

                # Check if this frame has any data for the relevant tile.
                any_column = False
                for word_idx in range(bits_info.bits.words):
                    if word_idx + bits_info.bits.offset in bitdata[frame][0]:
                        any_column = True
                        break

                if not any_column:
                    continue

                tiles_checked.add((bits_info.tile, bits_info.block_type))

                for fasm_line in self.find_features_in_tile(
                        bits_info.tile, bits_info.block_type, bits_info.bits,
                        solved_bitdata, bitdata, verbose=verbose):
                    if fasm_line not in emitted_features:
                        emitted_features.add(fasm_line)
                        yield fasm_line

            remaining_bits = bitdata[frame][1]
            if frame in solved_bitdata:
                remaining_bits -= solved_bitdata[frame]

            if len(remaining_bits) > 0 and verbose:
                # Some bits were not decoded, add warning and annotations to
                # FASM.
                yield fasm.FasmLine(
                    set_feature=None,
                    annotations=None,
                    comment=" In frame 0x{:08x} {} bits were not converted.".
                    format(
                        frame,
                        len(remaining_bits),
                    ))

                for bit in sorted(remaining_bits):
                    frame_offset = frame % bitstream.FRAME_ALIGNMENT
                    aligned_frame = frame - frame_offset
                    wordidx = bit // bitstream.WORD_SIZE_BITS
                    bitidx = bit % bitstream.WORD_SIZE_BITS

                    annotations = []
                    annotations.append(
                        fasm.Annotation(
                            'unknown_bit', '{:08x}_{}_{}'.format(
                                frame, wordidx, bitidx)))
                    annotations.append(
                        fasm.Annotation(
                            'unknown_segment',
                            '0x{:08x}'.format(aligned_frame)))
                    annotations.append(
                        fasm.Annotation(
                            'unknown_segbit', '{:02d}_{:02d}'.format(
                                frame_offset, bit)))
                    yield fasm.FasmLine(
                        set_feature=None,
                        annotations=tuple(annotations),
                        comment=None,
                    )

    def is_zero_feature(self, feature):
        parts = feature.split('.')
        tile = parts[0]
        gridinfo = self.grid.gridinfo_at_tilename(tile)
        feature = '.'.join(parts[1:])

        db_k = '%s.%s' % (gridinfo.tile_type, feature)
        segbits = self.grid.get_tile_segbits_at_tilename(tile)
        any_bits = False
        for block_type, bit in segbits.feature_to_bits(gridinfo.bits, db_k):
            if bit.isset:
                any_bits = True
                break

        return not any_bits



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
from collections import namedtuple
from prjxray import bitstream
from prjxray.grid_types import BlockType
from prjxray.util import OpenSafeFile
import enum


class PsuedoPipType(enum.Enum):
    ALWAYS = 'always'
    DEFAULT = 'default'
    HINT = 'hint'


def read_ppips(f):
    ppips = {}

    for l in f:
        l = l.strip()
        if not l:
            continue

        feature, ppip_type = l.split(' ')

        ppips[feature] = PsuedoPipType(ppip_type)

    return ppips


Bit = namedtuple('Bit', 'word_column word_bit isset')


def parsebit(val):
    '''Return "!012_23" => (12, 23, False)'''
    isset = True
    # Default is 0. Skip explicit call outs
    if val[0] == '!':
        isset = False
        val = val[1:]
    # 28_05 => 28, 05
    parts = val.split('_')
    assert len(parts) == 2, val
    seg_word_column, word_bit_n = parts

    return Bit(
        word_column=int(seg_word_column),
        word_bit=int(word_bit_n),
        isset=isset,
    )


def read_segbits(f):
    segbits = {}

    for l in f:
        # CLBLM_L.SLICEL_X1.ALUT.INIT[10] 29_14
        l = l.strip()

        if not l:
            continue

        parts = l.split(' ')

        assert len(parts) > 1, l

        segbits[parts[0]] = [parsebit(val) for val in parts[1:]]

    return segbits


class TileSegbits(object):
    def __init__(self, tile_db):
        self.segbits = {}
        self.ppips = {}
        self.feature_addresses = {}

        if tile_db.ppips is not None:
            with OpenSafeFile(tile_db.ppips) as f:
                self.ppips = read_ppips(f)

        if tile_db.segbits is not None:
            with OpenSafeFile(tile_db.segbits) as f:
                self.segbits[BlockType.CLB_IO_CLK] = read_segbits(f)

        if tile_db.block_ram_segbits is not None:
            with OpenSafeFile(tile_db.block_ram_segbits) as f:
                self.segbits[BlockType.BLOCK_RAM] = read_segbits(f)

        for block_type in self.segbits:
            for feature in self.segbits[block_type]:
                sidx = feature.rfind('[')
                eidx = feature.rfind(']')

                if sidx != -1:
                    assert eidx != -1

                    base_feature = feature[:sidx]

                    if base_feature not in self.feature_addresses:
                        self.feature_addresses[base_feature] = {}

                    self.feature_addresses[base_feature][int(
                        feature[sidx + 1:eidx])] = (block_type, feature)

    def match_bitdata(self, block_type, bits, bitdata, match_filter=None):
        """ Return matching features for tile bits data (grid.Bits) and bitdata.

        See bitstream.load_bitdata for details on bitdata structure.

        """

        if block_type not in self.segbits:
            return

        for feature, segbit in self.segbits[block_type].items():
            match = True
            skip = False
            for query_bit in segbit:
                if match_filter is not None and not match_filter(block_type,
                                                                 query_bit):
                    skip = True
                    break

                frame = bits.base_address + query_bit.word_column
                bitidx = bits.offset * bitstream.WORD_SIZE_BITS + query_bit.word_bit

                if frame not in bitdata:
                    match = not query_bit.isset
                    if match:
                        continue
                    else:
                        break

                found_bit = bitidx in bitdata[frame][1]
                match = found_bit == query_bit.isset

                if not match:
                    break

            if not match or skip:
                continue

            def inner():
                for query_bit in segbit:
                    if query_bit.isset:
                        frame = bits.base_address + query_bit.word_column
                        bitidx = bits.offset * bitstream.WORD_SIZE_BITS + query_bit.word_bit
                        yield (frame, bitidx)

            yield (tuple(inner()), feature)

    def map_bit_to_frame(self, block_type, bits, bit):
        """ Convert bit from segbit to frame data. """
        return Bit(
            word_column=bits.base_address + bit.word_column,
            word_bit=bits.offset * bitstream.WORD_SIZE_BITS + bit.word_bit,
            isset=bit.isset,
        )

    def feature_to_bits(self, bits_map, feature, address=0):
        if feature in self.ppips:
            return

        for block_type in self.segbits:
            if address == 0 and feature in self.segbits[block_type]:
                for bit in self.segbits[block_type][feature]:
                    yield block_type, self.map_bit_to_frame(
                        block_type, bits_map[block_type], bit)
                return

        block_type, feature = self.feature_addresses[feature][address]
        for bit in self.segbits[block_type][feature]:
            yield block_type, self.map_bit_to_frame(
                block_type, bits_map[block_type], bit)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import re
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database
from prjxray.grid_types import GridLoc

GTX_INT_Y_RE = re.compile("GTX_INT_INTERFACE.*X[0-9]+Y([0-9]+)")


def get_gtx_int_tile(clock_region, grid):
    for tile_name in sorted(grid.tiles()):
        if not tile_name.startswith("GTX_INT_INTERFACE"):
            continue

        loc = grid.loc_of_tilename(tile_name)

        left_gridinfo = grid.gridinfo_at_loc(
            GridLoc(loc.grid_x - 1, loc.grid_y))
        right_gridinfo = grid.gridinfo_at_loc(
            GridLoc(loc.grid_x + 1, loc.grid_y))

        if left_gridinfo.tile_type in ["INT_L", "INT_R"]:
            cmt = left_gridinfo.clock_region
        elif right_gridinfo.tile_type in ["INT_L", "INT_R"]:
            cmt = right_gridinfo.clock_region
        else:
            assert False

        gridinfo = grid.gridinfo_at_loc(loc)

        m = GTX_INT_Y_RE.match(tile_name)

        assert m

        int_y = int(m.group(1))

        if clock_region == cmt and int_y % 50 == 26:
            return tile_name


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['GTXE2_COMMON']:
                gtx_int_tile = get_gtx_int_tile(gridinfo.clock_region, grid)

                yield gtx_int_tile, site_name


def write_params(params):
    pinstr = 'tile,val,site\n'
    for tile, (site, val) in sorted(params.items()):
        pinstr += '%s,%s,%s\n' % (tile, val, site)
    open('params.csv', 'w').write(pinstr)


def run():
    print('''
module top();
    ''')

    params = {}

    sites = list(gen_sites())
    for gtx_int_tile, site_name in sites:
        isone = random.randint(0, 1)

        params[gtx_int_tile] = (site_name, isone)

        print(
            '''
wire QPLLLOCKEN_{site};

(* KEEP, DONT_TOUCH *)
LUT1 lut_{site} (.O(QPLLLOCKEN_{site}));

(* KEEP, DONT_TOUCH, LOC = "{site}" *)
GTXE2_COMMON gtxe2_common_{site} (
    .QPLLLOCKEN(QPLLLOCKEN_{site})
);'''.format(site=site_name))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import argparse

from migen import *

from litex.boards.platforms import arty
from litex.build.xilinx.vivado import vivado_build_args, vivado_build_argdict

from litex.soc.cores.clock import *
from litex.soc.integration.soc_sdram import *
from litex.soc.integration.builder import *

from litedram.modules import MT41K128M16
from litedram.phy import s7ddrphy

# CRG ----------------------------------------------------------------------------------------------


class _CRG(Module):
    def __init__(self, platform, sys_clk_freq):
        self.clock_domains.cd_sys = ClockDomain()
        self.clock_domains.cd_sys4x = ClockDomain(reset_less=True)
        self.clock_domains.cd_sys4x_dqs = ClockDomain(reset_less=True)
        self.clock_domains.cd_clk200 = ClockDomain()
        self.clock_domains.cd_eth = ClockDomain()

        # # #
        pll_clkin = Signal()
        self.specials += Instance(
            "BUFG", i_I=platform.request("clk100"), o_O=pll_clkin)
        self.submodules.pll = pll = S7PLL(speedgrade=-1)
        self.comb += pll.reset.eq(~platform.request("cpu_reset"))
        pll.register_clkin(pll_clkin, 100e6)
        pll.create_clkout(self.cd_sys, sys_clk_freq)
        pll.create_clkout(self.cd_sys4x, 4 * sys_clk_freq)
        pll.create_clkout(self.cd_sys4x_dqs, 4 * sys_clk_freq, phase=90)
        pll.create_clkout(self.cd_clk200, 200e6)
        self.submodules.idelayctrl = S7IDELAYCTRL(self.cd_clk200)


# BaseSoC ------------------------------------------------------------------------------------------


class MinSoC(SoCSDRAM):
    def __init__(
            self, sys_clk_freq=int(50e6), integrated_rom_size=0x8000,
            **kwargs):
        platform = arty.Platform()

        # SoCSDRAM ---------------------------------------------------------------------------------
        SoCSDRAM.__init__(
            self,
            platform,
            clk_freq=sys_clk_freq,
            integrated_rom_size=integrated_rom_size,
            integrated_sram_size=0x8000,
            cpu_variant="lite",
            **kwargs)

        # CRG --------------------------------------------------------------------------------------
        self.submodules.crg = _CRG(platform, sys_clk_freq)

        # DDR3 SDRAM -------------------------------------------------------------------------------
        if not self.integrated_main_ram_size:
            self.submodules.ddrphy = s7ddrphy.A7DDRPHY(
                platform.request("ddram"),
                memtype="DDR3",
                nphases=4,
                sys_clk_freq=sys_clk_freq)
            self.add_csr("ddrphy")
            sdram_module = MT41K128M16(sys_clk_freq, "1:4")
            self.register_sdram(
                self.ddrphy,
                geom_settings=sdram_module.geom_settings,
                timing_settings=sdram_module.timing_settings)


# Build --------------------------------------------------------------------------------------------


def main():
    parser = argparse.ArgumentParser(description="LiteX SoC on Arty")
    builder_args(parser)
    soc_sdram_args(parser)
    vivado_build_args(parser)
    args = parser.parse_args()

    cls = MinSoC
    soc = cls(**soc_sdram_argdict(args))
    builder = Builder(soc, **builder_argdict(args))
    builder.build(**vivado_build_argdict(args))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker
from prjxray import verilog


def add_enum_bits(segmk, params, key, options):
    for opt in options:
        segmk.add_site_tag(
            params['tile'], '{}_{}'.format(key, opt), params[key] == opt)


def output_integer_tags(segmk, params, key, invert=False):
    tile = params['tile']
    bits = verilog.parse_bitstr(params[key])
    for bit, tag_val in enumerate(bits):
        if not invert:
            segmk.add_tile_tag(
                tile, "{}[{}]".format(key,
                                      len(bits) - bit - 1), tag_val)
        else:
            segmk.add_tile_tag(
                tile, "Z{}[{}]".format(key,
                                       len(bits) - bit - 1),
                0 if tag_val else 1)


def main():
    segmk = Segmaker("design.bits")

    print("Loading tags")
    with open('params.json') as f:
        params = json.load(f)

    for tile_param in params:
        if tile_param['EN_SYN'] and tile_param['DATA_WIDTH'] == 4:
            output_integer_tags(
                segmk, tile_param, 'ALMOST_EMPTY_OFFSET', invert=True)
            output_integer_tags(
                segmk, tile_param, 'ALMOST_FULL_OFFSET', invert=True)

        for param in ('EN_SYN', 'FIRST_WORD_FALL_THROUGH'):
            segmk.add_tile_tag(tile_param['tile'], param, tile_param[param])

    segmk.compile()
    segmk.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        if gridinfo.tile_type != 'CFG_CENTER_MID':
            continue

        sites = {}
        for site_name, site_type in gridinfo.sites.items():
            if site_type not in sites:
                sites[site_type] = []
            sites[site_type].append(site_name)

        for site_type in sites:
            sites[site_type].sort()

        int_grid_x = loc.grid_x + 3
        int_tile_type = 'INT_L'

        int_tile_locs = []
        for dy in range(-9, 12):
            # Skip the VBREAK tile.
            if dy != 6:
                int_tile_locs.append((int_grid_x, loc.grid_y + dy), )

        int_tiles = []
        for int_tile_loc in int_tile_locs:
            int_gridinfo = grid.gridinfo_at_loc(int_tile_loc)
            assert int_gridinfo.tile_type == int_tile_type, (
                int_gridinfo.tile_type, int_tile_type)

            int_tiles.append(grid.tilename_at_loc(int_tile_loc))

        yield tile_name, sites, int_tiles


def write_params(params):
    pinstr = 'tile,val\n'
    for tile, (val) in sorted(params.items()):
        pinstr += '%s,%s\n' % (tile, val)
    open('params.csv', 'w').write(pinstr)


def run():
    print('''
module top();
    ''')

    sites = list(gen_sites())

    # Only on CFG_CENTER_MID expected.
    assert len(sites) == 1
    tile_name, sites, int_tiles = sites[0]

    assert len(sites['ICAP']) == 2, len(sites['ICAP'])

    # int_tiles[6]:
    # IMUX43 -> ICAP1_I31 = 0
    # IMUX42 -> ICAP1_I30 = toggle 0/1
    # int_tiles[7]:
    # IMUX43 -> ICAP1_I15 = 0
    # IMUX42 -> ICAP1_I14 = toggle 0/1
    # int_tiles[8]:
    # IMUX43 -> ICAP1_CSIB = 0
    # IMUX42 -> ICAP1_RDWRB = toggle 0/1

    ICAP1_I30 = random.randint(0, 1)
    ICAP1_I14 = random.randint(0, 1)
    ICAP1_RDWRB = random.randint(0, 1)

    params = {}
    params[int_tiles[6]] = ICAP1_I30
    params[int_tiles[7]] = ICAP1_I14
    params[int_tiles[8]] = ICAP1_RDWRB

    print(
        """
    wire [31:0] icap_i_{site};
    wire icap_rdwrd_{site};
    wire icap_csib_{site};

    assign icap_i_{site}[31] = 0;
    assign icap_i_{site}[30] = {ICAP1_I30};

    assign icap_i_{site}[15] = 0;
    assign icap_i_{site}[14] = {ICAP1_I14};

    assign icap_csib_{site} = 0;
    assign icap_rdwrb_{site} = {ICAP1_RDWRB};

    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    ICAPE2 icap_{site} (
        .I(icap_i_{site}),
        .RDWRB(icap_rdwrb_{site}),
        .CSIB(icap_csib_{site})
    );
    """.format(
            site=sites['ICAP'][1],
            ICAP1_I30=ICAP1_I30,
            ICAP1_I14=ICAP1_I14,
            ICAP1_RDWRB=ICAP1_RDWRB))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker
from prjxray import segmaker
from prjxray import verilog
import os
import json
import csv


def bitfilter(frame, word):
    if frame < 38:
        return False

    return True


def mk_drive_opt(iostandard, drive):
    if drive is None:
        drive = '_FIXED'
    return '{}.DRIVE.I{}'.format(iostandard, drive)


def drives_for_iostandard(iostandard):
    if iostandard in ['LVTTL', 'LVCMOS18']:
        drives = [4, 8, 12, 16, 24]
    elif iostandard == 'LVCMOS12':
        drives = [4, 8, 12]
    elif iostandard in ['SSTL135', 'SSTL15']:
        return ['_FIXED']
    else:
        drives = [4, 8, 12, 16]

    return drives


STEPDOWN_IOSTANDARDS = [
    'LVCMOS12', 'LVCMOS15', 'LVCMOS18', 'SSTL135', 'SSTL15'
]
IBUF_LOW_PWR_SUPPORTED = ['SSTL135', 'SSTL15', 'LVDS_25', 'TMDS_33']

ONLY_DIFF_IOSTANDARDS = ['LVDS_25', 'TMDS_33']


def main():
    # Create map of iobank -> sites
    iobanks = {}
    site_to_iobank = {}
    iobank_iostandards = {}
    with open(os.path.join(os.getenv('FUZDIR'), 'build', 'iobanks.txt')) as f:
        for l in f:
            iob_site, iobank = l.strip().split(',')
            iobank = int(iobank)

            if iobank not in iobanks:
                iobanks[iobank] = set()

            iobanks[iobank].add(iob_site)
            assert iob_site not in site_to_iobank
            site_to_iobank[iob_site] = iobank

    for iobank in iobanks:
        iobank_iostandards[iobank] = set()

    # Load a list of PUDC_B pin function tiles. They are configured differently
    # by the vendor tools so need to be skipped
    pudc_tiles = set()
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'pudc_sites.csv')) as f:
        for l in csv.DictReader(f):
            pudc_tiles.add(l["tile"])

    print("Loading tags")
    segmk = Segmaker("design.bits")
    '''
    port,site,tile,pin,slew,drive,pulltype
    di[0],IOB_X0Y107,LIOB33_X0Y107,A21,PULLDOWN
    di[10],IOB_X0Y147,LIOB33_X0Y147,F14,PULLUP
    '''
    with open('params.json', 'r') as f:
        design = json.load(f)

        diff_pairs = set()
        for d in design['tiles']:
            iostandard = verilog.unquote(d['IOSTANDARD'])
            if iostandard.startswith('DIFF_'):
                diff_pairs.add(d['pair_site'])

        for d in design['tiles']:
            site = d['site']
            tile = d['tile']

            if tile in pudc_tiles:
                continue

            if site in diff_pairs:
                continue

            iostandard = verilog.unquote(d['IOSTANDARD'])
            if iostandard.startswith('DIFF_'):
                iostandard = iostandard[5:]

            iobank_iostandards[site_to_iobank[site]].add(iostandard)

            segmk.add_site_tag(
                site, '_'.join(STEPDOWN_IOSTANDARDS) + '.STEPDOWN',
                iostandard in STEPDOWN_IOSTANDARDS)

            if 'IN_TERM' in d:
                segmaker.add_site_group_zero(
                    segmk, site, 'IN_TERM.', [
                        'NONE', 'UNTUNED_SPLIT_40', 'UNTUNED_SPLIT_50',
                        'UNTUNED_SPLIT_60'
                    ], 'NONE', d['IN_TERM'])

            only_diff_io = iostandard in ONLY_DIFF_IOSTANDARDS

            if d['type'] is None:
                segmk.add_site_tag(site, 'INOUT', 0)
                segmk.add_site_tag(site, '{}.IN_USE'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.IN'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.OUT'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.IN_ONLY'.format(iostandard), 0)
            elif d['type'] == 'IBUF':
                segmk.add_site_tag(site, 'INOUT', 0)
                segmk.add_site_tag(site, '{}.IN_USE'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.IN'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.IN_DIFF'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.OUT'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.IN_ONLY'.format(iostandard), 1)
                segmk.add_tile_tag(d['tile'], 'IN_DIFF', 0)

                if iostandard in IBUF_LOW_PWR_SUPPORTED:
                    segmk.add_site_tag(site, 'IBUF_LOW_PWR', d['IBUF_LOW_PWR'])
                    segmk.add_site_tag(
                        site, 'ZIBUF_LOW_PWR', 1 ^ d['IBUF_LOW_PWR'])
            elif d['type'] == 'IBUFDS':
                segmk.add_site_tag(site, 'INOUT', 0)
                segmk.add_site_tag(site, '{}.IN_USE'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.IN'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.IN_DIFF'.format(iostandard), 1)
                segmk.add_site_tag(
                    d['pair_site'], '{}.IN_DIFF'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.OUT'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.IN_ONLY'.format(iostandard), 1)
                segmk.add_tile_tag(d['tile'], 'IN_DIFF', 1)

                if iostandard in IBUF_LOW_PWR_SUPPORTED:
                    segmk.add_tile_tag(
                        tile, 'DIFF.IBUF_LOW_PWR', d['IBUF_LOW_PWR'])
                    segmk.add_tile_tag(
                        tile, 'DIFF.ZIBUF_LOW_PWR', 1 ^ d['IBUF_LOW_PWR'])
            elif d['type'] == 'OBUF':
                segmk.add_site_tag(site, 'INOUT', 0)
                segmk.add_site_tag(site, '{}.IN_USE'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.IN'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.OUT'.format(iostandard), 1)
                segmk.add_tile_tag(d['tile'], 'OUT_DIFF', 0)
            elif d['type'] == 'OBUFDS':
                segmk.add_site_tag(site, 'INOUT', 0)
                segmk.add_site_tag(site, '{}.IN_USE'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.IN'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.OUT'.format(iostandard), 1)
                segmk.add_tile_tag(
                    d['tile'], 'OUT_DIFF', 1 and not only_diff_io)
                segmk.add_tile_tag(d['tile'], 'OUT_TDIFF', 0)
            elif d['type'] == 'OBUFTDS':
                segmk.add_site_tag(site, 'INOUT', 0)
                segmk.add_site_tag(site, '{}.IN_USE'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.IN'.format(iostandard), 0)
                segmk.add_site_tag(site, '{}.OUT'.format(iostandard), 1)
                segmk.add_tile_tag(
                    d['tile'], 'OUT_DIFF', 1 and not only_diff_io)
                segmk.add_tile_tag(
                    d['tile'], 'OUT_TDIFF', 1 and not only_diff_io)
            elif d['type'] == 'IOBUF_INTERMDISABLE':
                segmk.add_site_tag(site, 'INOUT', 1)
                segmk.add_site_tag(site, '{}.IN_USE'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.IN'.format(iostandard), 1)
                segmk.add_site_tag(site, '{}.OUT'.format(iostandard), 1)

            if d['type'] is not None:
                segmaker.add_site_group_zero(
                    segmk, site, "PULLTYPE.",
                    ("NONE", "KEEPER", "PULLDOWN", "PULLUP"), "PULLDOWN",
                    verilog.unquote(d['PULLTYPE']))

            if d['type'] in [None, 'IBUF', 'IBUFDS']:
                continue

            drive_opts = set()
            for opt in ("LVCMOS25", "LVCMOS33", "LVCMOS18", "LVCMOS15",
                        "LVCMOS12", 'LVTTL'):
                for drive_opt in ("4", "8", "12", "16", "24"):
                    if drive_opt == "16" and opt == "LVCMOS12":
                        continue

                    if drive_opt == "24" and opt not in ["LVCMOS18", 'LVTTL']:
                        continue

                    drive_opts.add(mk_drive_opt(opt, drive_opt))

            drive_opts.add(mk_drive_opt("SSTL135", None))
            drive_opts.add(mk_drive_opt("SSTL15", None))
            drive_opts.add(mk_drive_opt("LVDS_25", None))
            drive_opts.add(mk_drive_opt("TMDS_33", None))

            segmaker.add_site_group_zero(
                segmk, site, '', drive_opts, mk_drive_opt('LVCMOS25', '12'),
                mk_drive_opt(iostandard, d['DRIVE']))

            if d['SLEW']:
                for opt in ["SLOW", "FAST"]:
                    segmk.add_site_tag(
                        site, iostandard + ".SLEW." + opt,
                        opt == verilog.unquote(d['SLEW']))

            if 'ibufdisable_wire' in d:
                segmk.add_site_tag(
                    site, 'IBUFDISABLE.I', d['ibufdisable_wire'] != '0')

            if 'intermdisable_wire' in d:
                segmk.add_site_tag(
                    site, 'INTERMDISABLE.I', d['intermdisable_wire'] != '0')

    site_to_cmt = {}
    site_to_tile = {}
    tile_to_cmt = {}
    cmt_to_idelay = {}
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'cmt_regions.csv')) as f:
        for l in f:
            site, tile, cmt = l.strip().split(',')
            site_to_tile[site] = tile

            site_to_cmt[site] = cmt
            tile_to_cmt[tile] = cmt

            # Given IDELAYCTRL's are only located in HCLK_IOI3 tiles, and
            # there is only on HCLK_IOI3 tile per CMT, update
            # CMT -> IDELAYCTRL / tile map.
            if 'IDELAYCTRL' in site:
                assert cmt not in cmt_to_idelay
                cmt_to_idelay[cmt] = site, tile

    # For each IOBANK with an active VREF set the feature
    cmt_vref_active = set()
    with open('iobank_vref.csv') as f:
        for l in f:
            iobank, vref = l.strip().split(',')
            iobank = int(iobank)

            cmt = None
            for cmt_site in iobanks[iobank]:
                if cmt_site in site_to_cmt:
                    cmt = site_to_cmt[cmt_site]
                    break

            if cmt is None:
                continue

            cmt_vref_active.add(cmt)

            _, hclk_cmt_tile = cmt_to_idelay[cmt]

            opt = 'VREF.V_{:d}_MV'.format(int(float(vref) * 1000))
            segmk.add_tile_tag(hclk_cmt_tile, opt, 1)

    for iobank in iobank_iostandards:
        if len(iobank_iostandards[iobank]) == 0:
            continue

        for cmt_site in iobanks[iobank]:
            if cmt_site in site_to_cmt:
                cmt = site_to_cmt[cmt_site]
                break

        if cmt is None:
            continue

        _, hclk_cmt_tile = cmt_to_idelay[cmt]

        assert len(iobank_iostandards[iobank]) == 1, iobank_iostandards[iobank]

        iostandard = list(iobank_iostandards[iobank])[0]
        segmk.add_tile_tag(
            hclk_cmt_tile, 'STEPDOWN', iostandard in STEPDOWN_IOSTANDARDS)

        for only_diff_io in ONLY_DIFF_IOSTANDARDS:
            segmk.add_tile_tag(
                hclk_cmt_tile, '{}_IN_USE'.format(only_diff_io),
                iostandard == only_diff_io)

        segmk.add_tile_tag(
            hclk_cmt_tile, 'ONLY_DIFF_IN_USE',
            iostandard in ONLY_DIFF_IOSTANDARDS)

    # For IOBANK's with no active VREF, clear all VREF options.
    for cmt, (_, hclk_cmt_tile) in cmt_to_idelay.items():
        if cmt in cmt_vref_active:
            continue

        for vref in (
                .600,
                .675,
                .75,
                .90,
        ):
            opt = 'VREF.V_{:d}_MV'.format(int(vref * 1000))
            segmk.add_tile_tag(hclk_cmt_tile, opt, 0)

    segmk.compile(bitfilter=bitfilter)
    segmk.write(allow_empty=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
import json
random.seed(int(os.getenv("SEED"), 16))
from prjxray.db import Database
from prjxray import util
from prjxray import verilog


def gen_bram36(grid):
    for tile_name in grid.tiles():
        loc = grid.loc_of_tilename(tile_name)

        gridinfo = grid.gridinfo_at_loc(loc)

        found = False
        for site_name, site_type in gridinfo.sites.items():
            if site_type == 'RAMBFIFO36E1':
                found = True
                break

        if found:
            bram36_site_name = site_name
            for site_name, site_type in gridinfo.sites.items():
                if site_type == 'RAMB18E1':
                    bram18_site_name = site_name

                if site_type == 'FIFO18E1':
                    fifo18_site_name = site_name

            yield tile_name, bram36_site_name, bram18_site_name, fifo18_site_name


RAM_EXTENSION_OPTS = [
    "NONE",
    "LOWER",
    "UPPER",
]

BRAM36_WIDTHS = [1, 2]
BRAM36_TO_18_WIDTHS = {1: 1, 2: 1}


def main():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    print('''
module top();
    ''')

    params = []
    for tile_name, bram36_site_name, bram18_site_name, fifo18_site_name in gen_bram36(
            grid):
        bram36_ra_width = random.choice(BRAM36_WIDTHS)
        bram36_wa_width = random.choice(BRAM36_WIDTHS)
        bram36_rb_width = random.choice(BRAM36_WIDTHS)
        bram36_wb_width = random.choice(BRAM36_WIDTHS)

        bram18_ra_width = BRAM36_TO_18_WIDTHS[bram36_ra_width]
        bram18_wa_width = BRAM36_TO_18_WIDTHS[bram36_wa_width]
        bram18_rb_width = BRAM36_TO_18_WIDTHS[bram36_rb_width]
        bram18_wb_width = BRAM36_TO_18_WIDTHS[bram36_wb_width]

        if random.random() < .8:
            if bram36_ra_width == 1 and bram36_wa_width == 1:
                ram_extension_a = random.choice(RAM_EXTENSION_OPTS)
            else:
                ram_extension_a = 'NONE'

            if bram36_rb_width == 1 and bram36_wb_width == 1:
                ram_extension_b = random.choice(RAM_EXTENSION_OPTS)
            else:
                ram_extension_b = 'NONE'

            en_ecc_read = random.randint(0, 1)
            en_ecc_write = random.randint(0, 1)

            print(
                '''
                (* KEEP, DONT_TOUCH, LOC = "{site}" *)
                RAMB36E1 #(
                    .READ_WIDTH_A({bram36_ra_width}),
                    .WRITE_WIDTH_A({bram36_wa_width}),
                    .READ_WIDTH_B({bram36_rb_width}),
                    .WRITE_WIDTH_B({bram36_wb_width}),
                    .RAM_EXTENSION_A({ram_extension_a}),
                    .RAM_EXTENSION_B({ram_extension_b}),
                    .EN_ECC_READ({en_ecc_read}),
                    .EN_ECC_WRITE({en_ecc_write})
                    ) bram_{site} (
                        .CLKARDCLK(),
                        .CLKBWRCLK(),
                        .ENARDEN(),
                        .ENBWREN(),
                        .REGCEAREGCE(),
                        .REGCEB(),
                        .RSTRAMARSTRAM(),
                        .RSTRAMB(),
                        .RSTREGARSTREG(),
                        .RSTREGB(),
                        .ADDRARDADDR(),
                        .ADDRBWRADDR(),
                        .DIADI(),
                        .DIBDI(),
                        .DIPADIP(),
                        .DIPBDIP(),
                        .WEA(),
                        .WEBWE(),
                        .DOADO(),
                        .DOBDO(),
                        .DOPADOP(),
                        .DOPBDOP());
                '''.format(
                    site=bram36_site_name,
                    ram_extension_a=verilog.quote(ram_extension_a),
                    ram_extension_b=verilog.quote(ram_extension_b),
                    en_ecc_read=en_ecc_read,
                    en_ecc_write=en_ecc_write,
                    bram36_ra_width=bram36_ra_width,
                    bram36_wa_width=bram36_wa_width,
                    bram36_rb_width=bram36_rb_width,
                    bram36_wb_width=bram36_wb_width,
                ))

            params.append(
                {
                    'tile': tile_name,
                    'BRAM36_IN_USE': True,
                    'site': bram36_site_name,
                    'RAM_EXTENSION_A': ram_extension_a,
                    'RAM_EXTENSION_B': ram_extension_b,
                    'EN_ECC_READ': en_ecc_read,
                    'EN_ECC_WRITE': en_ecc_write,
                    'bram36_ra_width': bram36_ra_width,
                    'bram36_wa_width': bram36_wa_width,
                    'bram36_rb_width': bram36_rb_width,
                    'bram36_wb_width': bram36_wb_width,
                })
        else:
            print(
                '''
                (* KEEP, DONT_TOUCH, LOC = "{bram18}" *)
                RAMB18E1 #(
                    .READ_WIDTH_A({bram18_ra_width}),
                    .WRITE_WIDTH_A({bram18_wa_width}),
                    .READ_WIDTH_B({bram18_rb_width}),
                    .WRITE_WIDTH_B({bram18_wb_width})
                    ) bram_{bram18} (
                        .CLKARDCLK(),
                        .CLKBWRCLK(),
                        .ENARDEN(),
                        .ENBWREN(),
                        .REGCEAREGCE(),
                        .REGCEB(),
                        .RSTRAMARSTRAM(),
                        .RSTRAMB(),
                        .RSTREGARSTREG(),
                        .RSTREGB(),
                        .ADDRARDADDR(),
                        .ADDRBWRADDR(),
                        .DIADI(),
                        .DIBDI(),
                        .DIPADIP(),
                        .DIPBDIP(),
                        .WEA(),
                        .WEBWE(),
                        .DOADO(),
                        .DOBDO(),
                        .DOPADOP(),
                        .DOPBDOP());

                (* KEEP, DONT_TOUCH, LOC = "{fifo18}" *)
                RAMB18E1 #(
                    .READ_WIDTH_A({bram18_ra_width}),
                    .WRITE_WIDTH_A({bram18_wa_width}),
                    .READ_WIDTH_B({bram18_rb_width}),
                    .WRITE_WIDTH_B({bram18_wb_width})
                    ) bram_{fifo18} (
                        .CLKARDCLK(),
                        .CLKBWRCLK(),
                        .ENARDEN(),
                        .ENBWREN(),
                        .REGCEAREGCE(),
                        .REGCEB(),
                        .RSTRAMARSTRAM(),
                        .RSTRAMB(),
                        .RSTREGARSTREG(),
                        .RSTREGB(),
                        .ADDRARDADDR(),
                        .ADDRBWRADDR(),
                        .DIADI(),
                        .DIBDI(),
                        .DIPADIP(),
                        .DIPBDIP(),
                        .WEA(),
                        .WEBWE(),
                        .DOADO(),
                        .DOBDO(),
                        .DOPADOP(),
                        .DOPBDOP());
                '''.format(
                    bram18=bram18_site_name,
                    fifo18=fifo18_site_name,
                    bram18_ra_width=bram18_ra_width,
                    bram18_wa_width=bram18_wa_width,
                    bram18_rb_width=bram18_rb_width,
                    bram18_wb_width=bram18_wb_width,
                ))

            params.append(
                {
                    'tile': tile_name,
                    'BRAM36_IN_USE': False,
                    'site': bram36_site_name,
                    'bram36_ra_width': bram36_ra_width,
                    'bram36_wa_width': bram36_wa_width,
                    'bram36_rb_width': bram36_rb_width,
                    'bram36_wb_width': bram36_wb_width,
                })

    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(params, f, indent=2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys
import json
import re


def main():
    db_tiles = set()
    db_tile_prop = dict()
    db_tile_sites = dict()

    db_sites = set()
    db_site_prop = dict()
    db_site_tile = dict()
    db_site_bit = dict()

    def add_tile(tile):
        if tile not in db_tiles:
            db_tiles.add(tile)
            db_tile_prop[tile] = dict()
            db_tile_sites[tile] = list()

    def add_site(site):
        if site not in db_sites:
            db_sites.add(site)
            db_site_prop[site] = dict()

    with open("%s.txt" % sys.argv[1]) as f:
        for line in f:
            line = line.split()

            if line[0] == "TILEPROP":
                add_tile(line[1])
                db_tile_prop[line[1]][line[2]] = " ".join(line[3:])
                continue

            if line[0] == "TILESITE":
                add_tile(line[1])
                add_site(line[2])
                db_tile_sites[line[1]].append(line[2])
                db_site_tile[line[2]] = line[1]
                continue

            if line[0] == "SITEPROP":
                add_site(line[1])
                db_site_prop[line[1]][line[2]] = " ".join(line[3:])
                continue

            if line[0] == "SLICEBIT":
                db_site_bit[line[1]] = line[2]
                continue

            assert False

    print("Number of tiles: %d" % len(db_tiles))
    print("Number of sites: %d" % len(db_sites))
    print("Number of sites with bit: %d" % len(db_site_bit))

    database = dict()
    loc_to_tile = dict()

    database["device"] = sys.argv[2]

    database["tiles"] = dict()
    for tile in db_tiles:
        entry = dict()
        entry["props"] = db_tile_prop[tile]
        entry["sites"] = db_tile_sites[tile]
        database["tiles"][tile] = entry

        col = int(db_tile_prop[tile]["COLUMN"])
        row = int(db_tile_prop[tile]["ROW"])
        loc_to_tile[(col, row)] = tile

    database["sites"] = dict()
    for site in db_sites:
        entry = dict()
        entry["props"] = db_site_prop[site]
        entry["tile"] = db_site_tile[site]
        database["sites"][site] = entry

    for site, bit in db_site_bit.items():
        bit = bit.split("_")
        bit_type = int(bit[4][1:])
        bit_half = int(bit[5][1:])
        bit_row = int(bit[6][1:])
        bit_col = int(bit[7][1:])
        bit_word = int(bit[9][1:])
        assert len(bit) == 11

        for i in range(50):
            m = re.match("(.*)Y([0-9]+)", site)
            this_site = "%sY%d" % (m.group(1), int(m.group(2)) + i)

            tile = db_site_tile[this_site]

            word = bit_word + 2 * i
            if word >= 50: word += 1

            entry = dict()
            entry["BASE_FRAMEID"] = "0x%08x" % (
                (bit_type << 23) | (bit_half << 22) | (bit_row << 17) |
                (bit_col << 7))
            entry["FRAME_TYPE"] = bit_type
            entry["FRAME_HALF"] = bit_half
            entry["FRAME_ROW"] = bit_row
            entry["FRAME_COLUMN"] = bit_col
            entry["WORDS"] = [word, word + 1]

            database["tiles"][tile]["cfgcol"] = entry

            if database["tiles"][tile]["props"]["TILE_TYPE"] in ("CLBLL_L",
                                                                 "CLBLM_L"):
                col = int(db_tile_prop[tile]["COLUMN"])
                row = int(db_tile_prop[tile]["ROW"])
                right_tile = loc_to_tile[(col + 1, row)]

                database["tiles"][right_tile]["cfgcol"] = entry

            if database["tiles"][tile]["props"]["TILE_TYPE"] in ("CLBLL_R",
                                                                 "CLBLM_R"):
                col = int(db_tile_prop[tile]["COLUMN"])
                row = int(db_tile_prop[tile]["ROW"])
                left_tile = loc_to_tile[(col - 1, row)]

                database["tiles"][left_tile]["cfgcol"] = entry

    tile_cfgcol_count = 0
    cfgcols = set()

    for tile in db_tiles:
        if "cfgcol" in database["tiles"][tile]:
            cfgcols.add(database["tiles"][tile]["cfgcol"]["BASE_FRAMEID"])
            tile_cfgcol_count += 1

    print("Number of assigned columns: %d" % len(cfgcols))
    print("Number of tiles with assigned column: %d" % tile_cfgcol_count)

    with open("%s.json" % sys.argv[1], "w") as f:
        print(json.dumps(database, sort_keys=True, indent="\t"), file=f)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
FDCE Primitive: D Flip-Flop with Clock Enable and Asynchronous Clear
FDPE Primitive: D Flip-Flop with Clock Enable and Asynchronous Preset
FDRE Primitive: D Flip-Flop with Clock Enable and Synchronous Reset
FDSE Primitive: D Flip-Flop with Clock Enable and Synchronous Set
LDCE Primitive: Transparent Data Latch with Asynchronous Clear and Gate Enable
LDPE Primitive: Transparent Data Latch with Asynchronous Preset and Gate Enable
'''

from prims import *

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")


def ones(l):
    #return l + [x + '_1' for x in l]
    #return sorted(l + [x + '_1' for x in l])
    ret = []
    for x in l:
        ret.append(x)
        ret.append(x + '_1')
    return ret


def loadtop():
    '''
    i,prim,loc,bel
    0,FDPE,SLICE_X12Y100,C5FF
    1,FDPE,SLICE_X15Y100,A5FF
    2,FDPE_1,SLICE_X16Y100,B5FF
    3,LDCE_1,SLICE_X17Y100,BFF
    '''
    f = open('top.txt', 'r')
    f.readline()
    ret = {}
    for l in f:
        i, prim, loc, bel, init = l.split(",")
        i = int(i)
        init = int(init)
        ret[loc] = (i, prim, loc, bel, init)
    return ret


top = loadtop()


def vs2i(s):
    return {"1'b0": 0, "1'b1": 1}[s]


print("Loading tags from design.txt")
with open("design.txt", "r") as f:
    for line in f:
        '''
        puts $fp "$type $tile $grid_x $grid_y $ff $bel_type $used $usedstr"

        CLBLM_L CLBLM_L_X10Y137 30 13 SLICE_X13Y137/AFF REG_INIT 1 FDRE
        CLBLM_L CLBLM_L_X10Y137 30 13 SLICE_X12Y137/D5FF FF_INIT 0
        '''
        line = line.split()
        tile_type = line[0]
        tile_name = line[1]
        grid_x = line[2]
        grid_y = line[3]
        # Other code uses BEL name
        # SLICE_X12Y137/D5FF
        site_ff_name = line[4]
        site, ff_name = site_ff_name.split('/')
        ff_type = line[5]
        used = int(line[6])
        cel_prim = None
        cel_name = None
        if used:
            cel_name = line[7]
            # ex: FDCE
            cel_prim = line[8]
            # 1'b1
            # cinv = int(line[9][-1])
            cinv = int(line[9])
            init = vs2i(line[10])
            #init = int(line[10])

        # A B C D
        which = ff_name[0]
        # LUT6 vs LUT5 FF
        is5 = '5' in ff_name

        if used:
            segmk.add_site_tag(site, "%s.ZINI" % ff_name, 1 ^ init)

            # CLKINV turns out to be more complicated than origianlly thought
            if isff(cel_prim):
                segmk.add_site_tag(site, "CLKINV", cinv)
                segmk.add_site_tag(site, "NOCLKINV", 1 ^ cinv)
            else:
                segmk.add_site_tag(site, "CLKINV", 1 ^ cinv)
                segmk.add_site_tag(site, "NOCLKINV", cinv)

            # Synchronous vs asynchronous FF
            # Unlike most bits, shared between all CLB FFs
            segmk.add_site_tag(site, "FFSYNC", cel_prim in ('FDSE', 'FDRE'))

            # Latch bit
            # Only applies to LUT6 (non-5) FF's
            if not is5:
                segmk.add_site_tag(site, "LATCH", isl(cel_prim))
            '''
            On name:
            The primitives you listed have a control input to set the FF value to zero (clear/reset),
            the other three primitives have a control input that sets the FF value to one.
            Z => inversion
            '''
            segmk.add_site_tag(
                site, "%s.ZRST" % ff_name,
                cel_prim in ('FDRE', 'FDCE', 'LDCE'))

segmk.compile()
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Reduce tile types to prototypes that are always correct.
The dump-all generate.tcl dumps all instances of each tile type.  Some tiles
are missing wires.  reduce_tile_types.py generates the superset tile that
encompases all tiles of that type.  If it is not possible to generate a super
set tile, an error will be generated.

"""

import argparse
import prjxray.lib
import prjxray.node_lookup
import datetime
import subprocess
import os.path
import pyjson5 as json5
import progressbar
import multiprocessing
import os
import functools
import json
from prjxray.xjson import extract_numbers


def check_and_strip_prefix(name, prefix):
    assert name.startswith(prefix), repr((name, prefix))
    return name[len(prefix):]


def flatten_site_pins(tile, site, site_pins, site_pin_node_to_wires):
    def inner():
        for site_pin in site_pins:
            wires = tuple(site_pin_node_to_wires(tile, site_pin['node']))

            if len(wires) == 0:
                yield (
                    check_and_strip_prefix(site_pin['site_pin'], site + '/'),
                    None)
                continue

            assert len(wires) == 1, repr(wires)

            pin_info = {
                'wire': wires[0],
                'speed_model_index': site_pin['speed_model_index'],
            }

            yield (
                check_and_strip_prefix(site_pin['site_pin'], site + '/'),
                pin_info)

    return dict(inner())


def compare_sites_and_update(tile, sites, new_sites):
    for site_a, site_b in zip(sites, new_sites):
        assert site_a['type'] == site_b['type']
        assert site_a['site_pins'].keys() == site_b['site_pins'].keys()

        for site_pin in site_a['site_pins']:
            if site_a['site_pins'][site_pin] is not None and site_b[
                    'site_pins'][site_pin] is not None:
                assert site_a['site_pins'][site_pin] == site_b['site_pins'][
                    site_pin]
            elif site_a['site_pins'][site_pin] is None and site_b['site_pins'][
                    site_pin] is not None:
                site_a['site_pins'][site_pin] = site_b['site_pins'][site_pin]


def get_prototype_site(site):
    proto = {}
    proto['type'] = site['type']
    proto['site_pins'] = {}
    proto['site_pips'] = {}
    for site_pin in site['site_pins']:
        name = check_and_strip_prefix(site_pin['site_pin'], site['site'] + '/')

        proto['site_pins'][name] = {
            'direction': site_pin['direction'],
        }

    for site_pip in site['site_pips']:
        name = check_and_strip_prefix(site_pip['site_pip'], site['site'] + '/')

        proto['site_pips'][name] = {
            'to_pin': site_pip['to_pin'],
            'from_pin': site_pip['from_pin'],
        }

    return proto


def get_pips(tile, pips):
    proto_pips = {}

    for pip in pips:
        name = check_and_strip_prefix(pip['pip'], tile + '/')

        proto_pips[name] = {
            'src_wire':
            check_and_strip_prefix(pip['src_wire'], tile + '/')
            if pip['src_wire'] is not None else None,
            'dst_wire':
            check_and_strip_prefix(pip['dst_wire'], tile + '/')
            if pip['dst_wire'] is not None else None,
            'is_pseudo':
            pip['is_pseudo'],
            'is_directional':
            pip['is_directional'],
            'can_invert':
            pip['can_invert'],
            'speed_model_index':
            pip['speed_model_index'],
        }

    return proto_pips


def compare_and_update_pips(pips, new_pips):
    # Pip names are always the same, but sometimes the src_wire or dst_wire
    # may be missing.

    assert pips.keys() == new_pips.keys(), repr((pips.keys(), new_pips.keys()))
    for name in pips:
        if pips[name]['src_wire'] is not None and new_pips[name][
                'src_wire'] is not None:
            assert pips[name]['src_wire'] == new_pips[name]['src_wire'], repr(
                (
                    pips[name]['src_wire'],
                    new_pips[name]['src_wire'],
                ))
        elif pips[name]['src_wire'] is None and new_pips[name][
                'src_wire'] is not None:
            pips[name]['src_wire'] = new_pips[name]['src_wire']

        if pips[name]['dst_wire'] is not None and new_pips[name][
                'dst_wire'] is not None:
            assert pips[name]['dst_wire'] == new_pips[name]['dst_wire'], repr(
                (
                    pips[name]['dst_wire'],
                    new_pips[name]['dst_wire'],
                ))
        elif pips[name]['dst_wire'] is None and new_pips[name][
                'dst_wire'] is not None:
            pips[name]['dst_wire'] = new_pips[name]['dst_wire']

        for k in ['is_pseudo', 'is_directional', 'can_invert']:
            assert pips[name][k] == new_pips[name][k], (
                k, pips[name][k], new_pips[name][k])


def check_wires(wires, sites, pips):
    """ Verify that the wires generates from nodes are a superset of wires in
      sites and pips """
    if sites is not None:
        for site in sites:
            for wire_to_site_pin in site['site_pins'].values():
                if wire_to_site_pin is not None:
                    assert wire_to_site_pin['wire'] in wires, repr(
                        (wire_to_site_pin, wires))

    if pips is not None:
        for pip in pips.values():
            if pip['src_wire'] is not None:
                assert pip['src_wire'] in wires, repr((pip['src_wire'], wires))
            if pip['dst_wire'] is not None:
                assert pip['dst_wire'] in wires, repr((pip['dst_wire'], wires))


def get_sites(tile, site_pin_node_to_wires):
    for site in tile['sites']:
        min_x_coord, min_y_coord = prjxray.lib.find_origin_coordinate(
            site['site'], (site['site'] for site in tile['sites']))

        orig_site_name = site['site']
        coordinate = prjxray.lib.get_site_coordinate_from_name(orig_site_name)

        x_coord = coordinate.x_coord - min_x_coord
        y_coord = coordinate.y_coord - min_y_coord

        yield (
            {
                'name':
                'X{}Y{}'.format(x_coord, y_coord),
                'prefix':
                coordinate.prefix,
                'x_coord':
                x_coord,
                'y_coord':
                y_coord,
                'type':
                site['type'],
                'site_pins':
                dict(
                    flatten_site_pins(
                        tile['tile'], site['site'], site['site_pins'],
                        site_pin_node_to_wires)),
            })


def read_json5(fname, database_file):
    node_lookup = prjxray.node_lookup.NodeLookup(database_file)

    with open(fname) as f:
        tile = json5.load(f)

    def get_site_types():
        for site in tile['sites']:
            yield get_prototype_site(site)

    site_types = tuple(get_site_types())
    sites = tuple(get_sites(tile, node_lookup.site_pin_node_to_wires))
    pips = get_pips(tile['tile'], tile['pips'])

    def inner():
        for wire in tile['wires']:
            assert wire['wire'].startswith(tile['tile'] + '/')

            wire_speed_model_index = wire['speed_model_index']

            yield wire['wire'][len(tile['tile']) + 1:], wire_speed_model_index

    wires = {k: v for (k, v) in inner()}
    wires_from_nodes = set(node_lookup.wires_for_tile(tile['tile']))
    assert len(wires_from_nodes - wires.keys()) == 0, repr(
        (wires, wires_from_nodes))

    return fname, tile, site_types, sites, pips, wires


def compare_and_update_wires(wires, new_wires):
    for wire in new_wires:
        if wire not in wires:
            wires[wire] = new_wires
        else:
            assert wires[wire] == new_wires[wire]


def get_speed_model_indices(reduced_tile):
    """ Extracts the speed model indices for the data structure """

    speed_model_indices = set()

    for site in reduced_tile['sites']:
        for site_pin in site['site_pins'].keys():
            if site['site_pins'][site_pin] is None:
                continue

            speed_model_indices.add(
                'site_pin,{}'.format(
                    site['site_pins'][site_pin]['speed_model_index']))

    for pip in reduced_tile['pips'].keys():
        speed_model_indices.add(
            'pip,{}'.format(reduced_tile['pips'][pip]['speed_model_index']))

    for wire in reduced_tile['wires'].keys():
        speed_model_indices.add('wire,{}'.format(reduced_tile['wires'][wire]))

    return speed_model_indices


def annotate_pips_speed_model(pips, speed_data):
    """ Updates the pips with correct timing data """

    for pip_name, pip_data in pips.items():
        speed_model_index = pip_data['speed_model_index']

        pip_speed_data = speed_data[speed_model_index]
        assert pip_speed_data['resource_name'] == 'pip', (
            pip_speed_data['resource_name'], speed_model_index)

        pips[pip_name]['is_pass_transistor'] = pip_speed_data[
            'is_pass_transistor']
        pips[pip_name]['src_to_dst'] = {
            'delay': pip_speed_data.get('forward_delay', None),
            'in_cap': pip_speed_data.get('forward_in_cap', None),
            'res': pip_speed_data.get('forward_res', None),
        }
        pips[pip_name]['dst_to_src'] = {
            'delay': pip_speed_data.get('reverse_delay', None),
            'in_cap': pip_speed_data.get('reverse_in_cap', None),
            'res': pip_speed_data.get('reverse_res', None),
        }

        del pips[pip_name]['speed_model_index']


def annotate_site_pins_speed_model(site_pins, speed_data):
    """ Updates the site_pins with correct timing data """

    for site_pin_name, pin_data in site_pins.items():
        if pin_data is None:
            continue

        speed_model_index = pin_data['speed_model_index']

        pin_speed_data = speed_data[speed_model_index]
        assert pin_speed_data['resource_name'] == 'site_pin', (
            pin_speed_data['resource_name'], speed_model_index)

        site_pins[site_pin_name]['delay'] = pin_speed_data['delay']

        cap = pin_speed_data['cap']
        res = pin_speed_data['res']
        if cap != 'null':
            site_pins[site_pin_name]['cap'] = cap
        if res != 'null':
            site_pins[site_pin_name]['res'] = res

        del site_pins[site_pin_name]['speed_model_index']


def annotate_wires_speed_model(wires, speed_data):
    """ Updates the wires with correct timing data """

    for wire_name, wire_data in wires.items():
        speed_model_index = wire_data

        wire_speed_data = speed_data[speed_model_index]
        assert wire_speed_data['resource_name'] == 'wire', (
            wire_speed_data['resource_name'], speed_model_index)

        cap = wire_speed_data['cap']
        res = wire_speed_data['res']
        if cap != '0.000' or res != '0.000':
            wires[wire_name] = {
                'cap': cap,
                'res': res,
            }
        else:
            wires[wire_name] = None


def annotate_speed_model(tile_type, reduced_tile, root_dir):
    """ Updates the reduced tile with the correct speed information """

    speed_model_indices = get_speed_model_indices(reduced_tile)

    tmp_indices_file = os.path.join(
        root_dir, '{}_speed_index.tmp'.format(tile_type))

    with open(tmp_indices_file, "w") as f:
        for index in speed_model_indices:
            print(index, file=f)

    # Get vivado path
    vivado = os.getenv('XRAY_VIVADO')
    assert vivado is not None

    subprocess.check_call(
        "{} -mode batch -source get_speed_model.tcl -tclargs {}".format(
            vivado, tmp_indices_file),
        shell=True,
        stdout=subprocess.DEVNULL)

    with open(tmp_indices_file, "r") as f:
        speed_model_data = json5.load(f)

    for site in reduced_tile['sites']:
        annotate_site_pins_speed_model(site['site_pins'], speed_model_data)

    annotate_pips_speed_model(reduced_tile['pips'], speed_model_data)
    annotate_wires_speed_model(reduced_tile['wires'], speed_model_data)


def reduce_tile(pool, site_types, tile_type, tile_instances, database_file):
    sites = None
    pips = None
    wires = None

    with progressbar.ProgressBar(max_value=len(tile_instances)) as bar:
        chunksize = 1
        if len(tile_instances) < chunksize * 2:
            iter = map(
                lambda file: read_json5(file, database_file), tile_instances)
        else:
            print(
                '{} Using pool.imap_unordered'.format(datetime.datetime.now()))
            iter = pool.imap_unordered(
                functools.partial(read_json5, database_file=database_file),
                tile_instances,
                chunksize=chunksize,
            )

        for idx, (fname, tile, new_site_types, new_sites, new_pips,
                  new_wires) in enumerate(iter):
            bar.update(idx)

            assert tile['type'] == tile_type, repr((tile['tile'], tile_type))

            for site_type in new_site_types:
                if site_type['type'] in site_types:
                    prjxray.lib.compare_prototype_site(
                        site_type, site_types[site_type['type']])
                else:
                    site_types[site_type['type']] = site_type

            # Sites are expect to always be the same
            if sites is None:
                sites = new_sites
            else:
                compare_sites_and_update(tile['tile'], sites, new_sites)

            if pips is None:
                pips = new_pips
            else:
                compare_and_update_pips(pips, new_pips)

            if wires is None:
                wires = new_wires
            else:
                compare_and_update_wires(wires, new_wires)

            bar.update(idx + 1)

    check_wires(wires, sites, pips)

    return {
        'tile_type': tile_type,
        'sites': sites,
        'pips': pips,
        'wires': wires,
    }


def main():
    parser = argparse.ArgumentParser(
        description=
        "Reduces raw database dump into prototype tiles, grid, and connections."
    )
    parser.add_argument('--root_dir', required=True)
    parser.add_argument('--output_dir', required=True)
    parser.add_argument('--ignore_cache', action='store_true')

    args = parser.parse_args()

    print('{} Reading root.csv'.format(datetime.datetime.now()))
    tiles, nodes = prjxray.lib.read_root_csv(args.root_dir)

    print('{} Loading node<->wire mapping'.format(datetime.datetime.now()))
    database_file = os.path.join(args.output_dir, 'nodes.db')
    if os.path.exists(database_file) and not args.ignore_cache:
        node_lookup = prjxray.node_lookup.NodeLookup(database_file)
    else:
        node_lookup = prjxray.node_lookup.NodeLookup(database_file)
        node_lookup.build_database(nodes=nodes, tiles=tiles)

    site_types = {}

    processes = multiprocessing.cpu_count()
    print('Running {} processes'.format(processes))
    pool = multiprocessing.Pool(processes=processes)

    for tile_type in sorted(tiles.keys()):
        #for tile_type in ['CLBLL_L', 'CLBLL_R', 'CLBLM_L', 'CLBLM_R', 'INT_L', 'INT_L']:
        tile_type_file = os.path.join(
            args.output_dir, 'tile_type_{}.json'.format(tile_type))
        site_types = {}
        if os.path.exists(tile_type_file):
            print(
                '{} Skip reduced tile for {}'.format(
                    datetime.datetime.now(), tile_type))
            continue
        print(
            '{} Generating reduced tile for {}'.format(
                datetime.datetime.now(), tile_type))
        reduced_tile = reduce_tile(
            pool, site_types, tile_type, tiles[tile_type], database_file)

        annotate_speed_model(tile_type, reduced_tile, args.root_dir)

        for site_type in site_types:
            with open(os.path.join(
                    args.output_dir, 'tile_type_{}_site_type_{}.json'.format(
                        tile_type, site_types[site_type]['type'])), 'w') as f:
                json.dump(site_types[site_type], f, indent=2, sort_keys=True)

        reduced_tile['sites'] = sorted(
            reduced_tile['sites'],
            key=lambda site: extract_numbers(
                '{}_{}'.format(site['name'], site['prefix'])))

        with open(tile_type_file, 'w') as f:
            json.dump(reduced_tile, f, indent=2, sort_keys=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" IOB bits are more complicated than can be easily expressed to segmaker.
There are couple cases that need to be handled here:

- There are some bits that are always set for IN-only ports, but are cleared
  selectively for OUT and INOUT ports.
- There are bits per each IOSTANDARD, in addition to drive patterns.  These
  can be merged to provide unique "(IOSTANDARD, DRIVE)" bit sets.
"""
import argparse


def get_name(l):
    parts = l.strip().split(' ')
    return parts[0]


def get_site(l):
    return get_name(l).split('.')[1]


def parse_bits(l):
    parts = l.strip().split(' ')
    if parts[1] in ['<0', '<const0>', '<const1>']:
        return frozenset()
    else:
        return frozenset(parts[1:])


def filter_bits(site, bits):
    """ Seperate top and bottom bits.

    Some IOSTANDARD bits are tile wide, but really only apply to a half.
    It is hard to write a fuzzer for this, but it is easy to filter by site,
    and all bits appear to have a nice hard halve seperatation in the bitidx.
    """
    if site == 'IOB_Y0':
        min_bitidx = 64
        max_bitidx = 127
    elif site == 'IOB_Y1':
        min_bitidx = 0
        max_bitidx = 63
    else:
        assert False, site

    def inner():
        for bit in bits:
            bitidx = int(bit.split('_')[1])

            if bitidx < min_bitidx or bitidx > max_bitidx:
                continue

            yield bit

    return frozenset(inner())


def process_features_sets(iostandard_lines):
    sites = {}

    for iostd_type, iostd_list in iostandard_lines.items():
        for iostd_line in iostd_list:
            feature = get_name(iostd_line)
            feature_parts = feature.split('.')
            site = get_site(iostd_line)
            iostandard = feature_parts[2]

            bits = parse_bits(iostd_line)
            bits = filter_bits(site, bits)

            key = (site, iostd_type)
            if key not in sites:
                sites[key] = {}

            group = feature_parts[3]
            if group not in sites[key]:
                sites[key][group] = {}

            if group in ['DRIVE', 'SLEW']:
                enum = feature_parts[4]
                sites[key][group][(iostandard, enum)] = bits
            elif group in ['IN', 'IN_DIFF', 'IN_ONLY', 'IN_USE', 'OUT',
                           'STEPDOWN', 'ZIBUF_LOW_PWR']:
                sites[key][group][(iostandard, None)] = bits
            else:
                assert False, group

    for site in sites:
        for iostandard, enum in sites[site]['DRIVE']:
            sites[site]['DRIVE'][(iostandard, enum)] |= sites[site]['OUT'][(
                iostandard, None)]

        for iostandard, enum in sites[site]['IN']:
            sites[site]['IN_ONLY'][(iostandard, enum)] -= sites[site]['IN'][(
                iostandard, enum)]

    common_bits = {}
    for site, iostd_type in sites:
        for group in sites[(site, iostd_type)]:
            if (site, group) not in common_bits:
                common_bits[(site, group)] = set()

            for bits in sites[(site, iostd_type)][group].values():
                common_bits[(site, group)] |= bits

    slew_in_drives = {}

    for site, iostd_type in sites:
        common_bits[(site, 'IN')] |= common_bits[(site, 'IN_DIFF')]
        common_bits[(site, 'IN_DIFF')] |= common_bits[(site, 'IN')]

        # Only DIFF IOSTANDARDS such as LVDS or TMDS do not have DRIVE,
        # STEPDOWN or SLEW features
        if iostd_type == "NORMAL":
            key = (site, iostd_type)
            common_bits[(site, 'DRIVE')] -= common_bits[(site, 'SLEW')]
            common_bits[(site, 'DRIVE')] -= common_bits[(site, 'STEPDOWN')]
            common_bits[(site, 'IN_ONLY')] |= common_bits[(site, 'DRIVE')]
            common_bits[(site, 'IN_ONLY')] -= common_bits[(site, 'STEPDOWN')]

            for iostandard, enum in sites[key]['DRIVE']:
                slew_in_drive = common_bits[
                    (site, 'SLEW')] & sites[key]['DRIVE'][(iostandard, enum)]
                if slew_in_drive:
                    if (key, iostandard) not in slew_in_drives:
                        slew_in_drives[(key, iostandard)] = set()

                    slew_in_drives[(key, iostandard)] |= slew_in_drive
                    sites[key]['DRIVE'][(iostandard, enum)] -= slew_in_drive

                sites[key]['DRIVE'][(iostandard,
                                     enum)] -= common_bits[(site, 'STEPDOWN')]

    for site, iostandard in slew_in_drives:
        for _, enum in sites[site]['SLEW']:
            sites[site]['SLEW'][(iostandard,
                                 enum)] |= slew_in_drives[(site, iostandard)]

    for site in sites:
        for iostandard, enum in sites[site]['DRIVE']:
            sites[site]['DRIVE'][(iostandard, enum)] |= sites[site]['IN_USE'][(
                iostandard, None)]

        for iostandard, enum in sites[site]['IN']:
            _, iostd_type = site
            if iostd_type == "ONLY_DIFF":
                sites[site]['IN_DIFF'][(iostandard, enum)] = \
                        sites[site]['IN'][(iostandard, enum)]
            elif sites[site]['IN_DIFF'][(iostandard, enum)]:
                sites[site]['IN_DIFF'][(iostandard, enum)] |= \
                        sites[site]['IN'][(iostandard, enum)]

    for site, iostd_type in sites:
        del sites[(site, iostd_type)]['IN_USE']
        if iostd_type == "NORMAL":
            del sites[(site, iostd_type)]['OUT']

    allow_zero = ['SLEW']

    common_groups = dict()
    for site, iostd_type in sites:
        if site not in common_groups:
            common_groups[site] = dict()

        key = (site, iostd_type)
        for group in sites[key]:
            if iostd_type == "ONLY_DIFF" and group == "IN":
                continue

            # Merge features that are identical.
            #
            # For example:
            #
            #  IOB33.IOB_Y1.LVCMOS15.IN 38_42 39_41
            #  IOB33.IOB_Y1.LVCMOS18.IN 38_42 39_41
            #
            # Must be grouped.
            for (iostandard, enum), bits in sites[key][group].items():
                if (bits, group) not in common_groups[site]:
                    common_groups[site][(bits, group)] = {
                        'IOSTANDARDS': set(),
                        'enums': set(),
                    }

                common_groups[site][(bits,
                                     group)]['IOSTANDARDS'].add(iostandard)
                if enum is not None:
                    common_groups[site][(bits, group)]['enums'].add(enum)

    visited_iostandards = list()
    for site, groups in common_groups.items():
        for (bits, group), v in groups.items():
            iostandards = v['IOSTANDARDS']
            enums = v['enums']

            # It happens that some features appear only in one of the IOB sites and not
            # in the other. This makes it hard to assign the correct features to the correct
            # site in the P&R toolchain.
            #
            # The following code makes sure that the same set of iostandards
            # (even if not really present at a site location) appears for each site
            for visited_iostandard, visited_group, visited_enums in visited_iostandards:
                same_enum = enums == visited_enums
                same_group = group == visited_group
                compatible_iostd = any(
                    x in iostandards for x in visited_iostandard)
                take_visited_iostd = len(visited_iostandard) > len(iostandards)
                if same_enum and same_group and compatible_iostd and take_visited_iostd:
                    iostandards = visited_iostandard
                    break

            visited_iostandards.append((iostandards, group, enums))

            iostandards_string = '_'.join(sorted(iostandards))

            if enums:
                feature = 'IOB33.{site}.{iostandards}.{group}.{enums}'.format(
                    site=site,
                    iostandards=iostandards_string,
                    group=group,
                    enums='_'.join(sorted(enums)),
                )
            else:
                feature = 'IOB33.{site}.{iostandards}.{group}'.format(
                    site=site,
                    iostandards=iostandards_string,
                    group=group,
                )

            if not bits and group not in allow_zero:
                continue

            neg_bits = frozenset(
                '!{}'.format(b) for b in (common_bits[(site, group)] - bits))
            print('{} {}'.format(feature, ' '.join(sorted(bits | neg_bits))))


def main():
    parser = argparse.ArgumentParser(
        description="Convert IOB rdb into good rdb."
        "")
    parser.add_argument('input_rdb')

    args = parser.parse_args()

    iostandard_lines = {
        "NORMAL": list(),
        "ONLY_DIFF": list(),
    }

    with open(args.input_rdb) as f:
        for l in f:
            if ('.SSTL' in l or '.LVCMOS' in l
                    or '.LVTTL' in l) and 'IOB_' in l:
                iostandard_lines["NORMAL"].append(l)
            elif ('.TMDS' in l or 'LVDS' in l):
                iostandard_lines["ONLY_DIFF"].append(l)
            else:
                print(l.strip())

    process_features_sets(iostandard_lines)


if __name__ == "__main__":
    main()
